/* acf.c -- access control file functions.
 *
 * $Id: acf.c,v 1.10 1998/10/16 22:05:39 ccarp Exp $
 * 
 * runas - execute a single command as a different user and/or group
 * Copyright (C) 1998 Chris Carpinello
 *
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 * Send all bugs, changes, and problems to the maintainer listed in
 * "runas --info" output.
 */

/* System headers */
#include <stdio.h>
#if (HAVE_STDLIB_H == 1)
#include <stdlib.h>
#endif /* HAVE_STDLIB_H */
#include <pwd.h>
#include <grp.h>
#include <sys/param.h>
#include <sys/stat.h> 
#include <sys/types.h> 
#if (HAVE_LIMITS_H == 1)
#include <limits.h>
#endif /* HAVE_LIMITS_H */
#include <ctype.h> 
#if (HAVE_MALLOC_H == 1)
#include <malloc.h>
#endif /* HAVE_MALLOC_H */
#include <string.h>
#if (HAVE_UNISTD_H == 1)
#include <unistd.h>
#endif /* HAVE_UNISTD_H */
#if (HAVE_SETNETGRENT == 1)
#if (HAVE_ARPA_NAMESER_H == 1)
#include <arpa/nameser.h>
#endif /* HAVE_ARPA_NAMESER_H */
#if (HAVE_NETDB_H == 1)
#include <netdb.h>
#endif /* HAVE_NETDB_H */
#endif /* HAVE_SETNETGRENT */

/* Local headers and macros */
#include "defines.h"
#include "config.h"
#include "acf.h"

/* ifdef USE_ACF acf.c */
#if (USE_ACF == TRUE)  /* Nothing to compile unless this is set */

/* File scope variables */
accessentry *accesscf = NULL;  /* the access control file list   */
#if (USE_SYSLOG == TRUE) 
char reason[MAXREASONSIZE];     /* reason for using runas command */
#endif /* USE_SYSLOG */

/* External variables */
#include "extern.h"

/* External function prototypes */ 
#include "proto.h"

/* Local definitions */
#define REMOVE 0
#define ADD    1

/* Local function prototypes */
PRIVATE void addToAccessList          __P((const char *user, accessentry *acfentry));
PRIVATE accessentry *findAcfEntry     __P((const char *)); 
PRIVATE accessentry *initAcfEntry     __P((uid_t, gid_t)); 
PRIVATE alist *initAcfUserEntry       __P((void));   
PRIVATE void parseAccessField         __P((char *, accessentry *, int, uid_t, gid_t));
PRIVATE void parseGroup               __P((const char *, accessentry *, uid_t, gid_t));
PRIVATE void parseNetgroup            __P((char *, accessentry *));
PRIVATE int removeFromAccessList      __P((const char *, accessentry *));  
PRIVATE void sortAccessField          __P((char *));

/* Functions */

/*
 * 1. Initialize access list.
 * 2. Iterate through the access list, copying positive entries
 *    to the positive access list and copying negative entries
 *    to the negative access list.
 * 3. Copy the positive access list to the beginning of the access list.
 * 4. If any negative access list entries exist, append them to the
 *    access list.
 *
 * char  *accesslist;                 access list to sort
 */
PRIVATE void sortAccessField(char *accesslist)
{
  char *token;
  char negativeaccesslist[MAX_ACCESS_ELEMENTSIZE]; /* users denied access  */
  char positiveaccesslist[MAX_ACCESS_ELEMENTSIZE]; /* users granted access */

  /*
   * 1. Initialize access list.
   */
  (void) bzero(negativeaccesslist, MAX_ACCESS_ELEMENTSIZE);
  (void) bzero(positiveaccesslist, MAX_ACCESS_ELEMENTSIZE);

  /*
   * 2. Iterate through the access list, copying positive entries
   *    to the positive access list and copying negative entries
   *    to the negative access list.
   */
  token = strtok(accesslist, ",");
  while (token != NULL) {
    if (strncmp(token, "-", 1) == 0) {
      if (negativeaccesslist[0] == (char)NULL)
	(void) strncpy(negativeaccesslist, token, MAXUSERNAMELEN);
      else {
	(void) strncat(negativeaccesslist, ",", 1);
	(void) strncat(negativeaccesslist, token, MAXUSERNAMELEN);
      }
    }
    else {
      if (positiveaccesslist[0] == (char)NULL)
	(void) strncpy(positiveaccesslist, token, MAXUSERNAMELEN);
      else {
	(void) strncat(positiveaccesslist, ",", 1);
	(void) strncat(positiveaccesslist, token, MAXUSERNAMELEN);
      }
    }    
    token = strtok(NULL, ",");
  }

  /*
   * 3. Copy the positive access list to the beginning of the access list.
   */
  (void) strncpy(accesslist, positiveaccesslist, MAX_ACCESS_ELEMENTLEN);

  /*
   * 4. If any negative access list entries exist, append them to the
   *    access list.
   */
  if (negativeaccesslist[0] != (char)NULL) {
    (void) strncat(accesslist, ",", 1);
    (void) strncat(accesslist, negativeaccesslist, MAX_ACCESS_ELEMENTLEN - strlen(accesslist) - 1);
  }
} /* sortAccessField */

#if (HAVE_SETNETGRENT == 1)
/*
 * 1. If the first character is not a minus sign, then set the 
 *    action to add all members of the the group to the acces list.
 * 2. Remove the plus or minus sign from the netgroup name. 
 * 3. Remove the @ delimiter from the netgroup name. 
 * 4. Rewind to the first netgroup entry.
 * 5. Iterate through the netgroup members and add or remove them from
 *    the access list accordingly.
 *
 * char *netgroup;          netgroup to parse               
 * accessentry *acfentry;   access control entry to modifiy 
 */
PRIVATE void parseNetgroup(char *netgroup, accessentry *acfentry)
{
  char *machinep[MAXHOSTNAMELEN+1];      /* netgroup machine info */
  char *userp[MAXUSERNAMESIZE];         /* netgroup user info    */
  char *domainp[MAXDOMAINNAMELEN+1];     /* netgroup domain info  */
  int action = REMOVE; /* remove user from access list by default */

  /* 
   * 1. If the first character is not a minus sign, then set the 
   *    action to add all members of the the group to the acces list.
   */
  if (strncmp(netgroup, "-", 1) != 0) {
    action = ADD;
  }

  /* 
   * 2. Remove the plus or minus sign from the netgroup name. 
   */
  if ((strncmp(netgroup, "+", 1) == 0) || (strncmp(netgroup, "-", 1) == 0)) {
    netgroup++;
  }

  /* 
   * 3. Remove the @ delimiter from the netgroup name. 
   */
  netgroup++; 

  /* 
   * 4. Rewind to the first netgroup entry.
   */
  (void) setnetgrent(netgroup); 

  /*
   * 5. Iterate through the netgroup members and add or remove them from
   *    the access list accordingly.
   */
  for (;;) {
    if (getnetgrent(machinep, userp, domainp) == 0)
	break;
    if (action == ADD) {
      (void) addToAccessList(*userp, acfentry);
    }
    else {
      (void) removeFromAccessList(*userp, acfentry);
    }
  }
} /* parseNetgroup */
#endif /* HAVE_SETNETGRENT */

/*
 * 1. If the first character is not a minus sign, then set the 
 *    action to add all members of the the group to the acces list.
 * 2. Remove the plus or minus sign from the group name. 
 * 3. Remove the % delimiter from the group name. 
 * 4. Verify that the group exists. 
 * 5. Iterate through the group members and add or remove them from
 *    the access list accordingly.
 * 6. Compare the user's base gid with the current access entry gid
 *    since getgrnam does not include users with a base group of 
 *    gr_nam.  If they are the same, then add or remove the user from
 *    the access list.
 *
 * const char *group;      user group
 * accessentry *acfentry;  access control entry
 * uid_t myuid;            user's uid
 * gid_t mygid;            user's gid                              
 */
PRIVATE void parseGroup(const char *group, 
			accessentry *acfentry, 
			uid_t myuid, 
			gid_t mygid)
{
  struct group *gr;    /* group information on gr_nam             */
  int action = REMOVE; /* remove user from access list by default */
  char **members;      /* list of all group members               */
  char newgroup[MAXGROUPNAMESIZE]; /* verified group name         */

  /* 
   * 1. If the first character is not a minus sign, then set the 
   *    action to add all members of the the group to the acces list.
   */
  if (strncmp(group, "-", 1) != 0) {
    action = ADD;
  }

  /* 
   * 2. Remove the plus or minus sign from the group name. 
   */
  if ((strncmp(group, "+", 1) == 0) || (strncmp(group, "-", 1) == 0)) {
    group++;
  }

  /* 
   * 3. Remove the % delimiter from the group name. 
   */
  group++;

  /* 
   * 4. Verify that the group exists. 
   */ 
  (void) bzero(newgroup, MAXGROUPNAMESIZE);
  (void) strncpy(newgroup, group, MAXGROUPNAMELEN);
  (void) validateGroup(newgroup);

  /* 
   * 5. Iterate through the group members and add or remove them from
   *    the access list accordingly.
   */
  gr = getgrnam(newgroup);
  members = gr->gr_mem;
  while (*members) {
    if (action == ADD) {
      (void) addToAccessList(*members, acfentry);
    }
    else {
      (void) removeFromAccessList(*members, acfentry);
    }
    members++;
  }

  /* 
   * 6. Compare the user's base gid with the current access entry gid
   *    since getgrnam does not include users with a base group of 
   *    gr_nam.  If they are the same, then add or remove the user from
   *    the access list.
   */ 
  if (gr->gr_gid == mygid) {
    if (action == ADD) {
      (void) addToAccessList((getpwuid(myuid))->pw_name, acfentry);
    } else {
      (void) removeFromAccessList((getpwuid(myuid))->pw_name, acfentry);
    }
  }
} /* parseGroup */

/*
 * 1. Only operate on a non-NULL access control list.
 * 2. Verify that the access control list has not been truncated
 *    because the entry in the access control file was larger than
 *    the buffer allocated.  If the access control list has been
 *    truncated, then remove the last entry from the access list.
 * 3. Sort the access list so all positive entries occur before
 *    the negative entries.
 * 4. Parse the access list containing groups, netgroups, usernames,
 *    and uids into usernames.
 *
 * char *accesslist;         access list to parse
 * accessentry *acfentry;    master access control structure
 * int lineoverflow;         buffer overflow occured in ACF_FILE
 * uid_t myuid;              uid of user invoking runas
 * gid_t mygid;              gid of user invoking runas
 */
PRIVATE void parseAccessField(char *accesslist, 
			      accessentry *acfentry, 
			      int lineoverflow, 
			      uid_t myuid, 
			      gid_t mygid)
{
  char *accesstoken;     /* users to add to access list            */
  char *delimiter;       /* strrchr pointer to character delimiter */
  size_t accesslist_len; /* length of accesslist                   */
  size_t delimiter_len;  /* length of accesslist                   */

  /*
   * 1. Only operate on a non-NULL access control list.
   */
  if (accesslist != NULL) {
    /*
     * 2. Verify that the access control list has not been truncated
     *    because the entry in the access control file was larger than
     *    the buffer allocated.  If the access control list has been
     *    truncated, then remove the last entry from the access list.
     */
    if (lineoverflow == TRUE) {
      delimiter = strrchr(accesslist, COMMA);
      if (delimiter == NULL) { /* one long bogus entry */
	accesslist = NULL;
      } else {
	/* remove data from beginning of incomplete access list entry */
        accesslist_len = strlen(accesslist); 
	delimiter_len = strlen(delimiter);
	accesslist[accesslist_len - delimiter_len] = (char)NULL;
      } /* if (delimiter == NULL) */
    } /* if (lineoverflow == TRUE) */
    if (accesslist != NULL) {

      /*
       * 3. Sort the access list so all positive entries occur before
       *    the negative entries.
       */
      (void) sortAccessField(accesslist); /* see function description */

      /* 
       * 4. Parse the access list containing groups, netgroups, usernames,
       *    and uids into usernames.
       */
      accesstoken = strtok(accesslist, ",");
      (void) cleanString(accesstoken); 
      for (;;) {
#if (HAVE_SETNETGRENT == 1)
	if (strstr(accesstoken, "@") != NULL) {
	  /* netgroup entry */
	  (void) parseNetgroup(accesstoken, acfentry);
	} else
#endif /* HAVE_SETNETGRENT */
	  if (strstr(accesstoken, "%") != NULL) { 
	    /* group entry */
	    (void) parseGroup(accesstoken, acfentry, myuid, mygid);
	  } else { 
	    /* user entry */
	    if (strncmp(accesstoken, "-", 1) == 0) {
	      /* don't pass the "-" thorugh */
	      (void) removeFromAccessList(++accesstoken, acfentry);
	    } else {
	      if (strncmp(accesstoken, "+", 1) == 0) {
		/* don't pass the "+" through */
		(void) addToAccessList(++accesstoken, acfentry);
	      } else {
		(void) addToAccessList(accesstoken, acfentry);
	      } /* strncmp(accessentry, "+", 1) */
	    }   /* strncmp(accessentry, "-", 1) */
	  }     /* strstr(accessentry, "%")     */
	/* get the next entry on the line */
	if ((accesstoken = strtok(NULL, ",")) == NULL)
	  break;
	(void) cleanString(accesstoken);
      } /* for (;;) */
    }   /* if (accesslist != NULL) */
  }     /* if (accesslist != NULL) */
} /* parseAccessField */

/*
 * 1. Open the access control file.
 * 2. Initialize the access control file entry list.
 * 3. Read in the access control file
 * 4. Parse the data into temporary buffers.
 * 5. Parse the access control list.
 * 6. Initialize the access control file entry.
 * 7. If the access control file exists but contains no data, then
 *    free the single allocated entry and return an error.
 * 8. Close the access control file.
 *
 * Returns -1 if an error occures while parsing the access control file.
 * Returns 0 if no error occured.
 *
 * uid_t myuid;                uid of user invoking runas
 * gid_t mygid;                gid of user invoking runas
 */
PUBLIC int parseAcf(uid_t myuid, gid_t mygid)
{
  accessentry *head = NULL; /* the start of ther access control list         */
  accessentry *prev = NULL; /* the previous entry in the access control list */
  FILE *acf = NULL;                /* file descriptor - access control file  */
  char tmpbuf[MAX_ACCESS_ENTRYSIZE];  /* buffer to read each acf line into    */
  char tmpuserlist[MAX_ACCESS_ELEMENTSIZE]; /* tmp user access list buffer    */
  char tmpargv[4];                 /* tmp arguments allowed buffer           */
  char tmpreason[4];               /* tmp reason needed buffer               */
  char tmprunuid[MAXUSERNAMESIZE];  /* tmp uid to runas buffer              */
  char tmprungid[MAXGROUPNAMESIZE]; /* tmp gid to runas buffer              */
  int acfentries = 0;              /* initial number of total acf entries    */
  int lineoverflow;                /* buffer too small for access list input */

  /* 
   * 1. Open the access control file.
   */
  acf = fopen(ACF_PATH, "r");
  if (acf == NULL) {
    perror(ACF_PATH);
    return(-1);
  }

  /*
   * 2. Initialize the access control file entry list.
   */
  head = accesscf = initAcfEntry(myuid, mygid); 

  /* Note: format of the access control file is:
   * [basename] [absolute pathname] [runuid] [rungid] [args] [reason] [access list] */

  (void) bzero(tmpbuf, sizeof(tmpbuf));

  /* 
   * 3. Read in the access control file
   */
  while (fgets(tmpbuf, MAX_ACCESS_ENTRYLEN, acf) != NULL) {
    /* probably should check for blank lines too */
    if (strncmp(tmpbuf,"#",1) != 0) { /* comment */
      (void) bzero(tmpuserlist, sizeof(tmpuserlist));
      (void) bzero(tmpargv, sizeof(tmpargv));
      (void) bzero(tmpreason, sizeof(tmpreason));
      (void) bzero(tmprunuid, sizeof(tmprunuid));
      (void) bzero(tmprungid, sizeof(tmprungid));

      lineoverflow = FALSE; /* buffer overflow during access list input */
      if (strstr(tmpbuf, NEWLINE) == NULL) {
	lineoverflow = TRUE;
      }

      /*
       * 4. Parse the data into temporary buffers.
       */
      (void) sscanf(tmpbuf, "%s %s %s %s %s %s %s", accesscf->mnemonic, accesscf->absolute, tmprunuid, tmprungid, tmpargv, tmpreason, tmpuserlist);
      accesscf->argv = convertToBoolean(tmpargv);
      accesscf->reason = convertToBoolean(tmpreason);
      /*
       * Note: if either user or group is "null" or "-1" then keep the
       * uid or gid value the same as the current uid or gid.
       */
      if ((strncmp(tmprunuid, "-1", 2) != 0) &&
	  (strncasecmp(tmprunuid, "null", 4) != 0)) {
	(void) strncpy(accesscf->runuid, tmprunuid, MAXUSERNAMELEN);
      } else {
	(void) snprintf(accesscf->runuid, MAXUSERNAMELEN, "%lu", (long unsigned)myuid);
      }
      (void) validateUser(accesscf->runuid);
      if ((strncmp(tmprungid, "-1", 2) != 0) &&
	  (strncasecmp(tmprungid, "null", 4) != 0)) {
	(void) strncpy(accesscf->rungid, tmprungid, MAXGROUPNAMELEN); 
      } else {
	(void) snprintf(accesscf->rungid, MAXGROUPNAMELEN, "%lu", (long unsigned)mygid);
      }
      (void) validateGroup(accesscf->rungid);
      /*
       * 5. Parse the access control list.
       */
      (void) parseAccessField(tmpuserlist, accesscf, lineoverflow, myuid, mygid);
      prev = accesscf;
      /* 
       * 6. Initialize the access control file entry.
       */
      accesscf->next = initAcfEntry(myuid, mygid); 
      accesscf = accesscf->next;
      acfentries++; /* total number of acf entries */
      /* Note: stop reading if special case is reached */
      if ((strncmp(prev->mnemonic, "*", 1) == 0) ||
          (strncmp(prev->absolute, "*", 1) == 0)) {
	if (fseek(acf, 0, SEEK_END) != 0) {
	  (void) esprintf(error, "%s: \"%s\", line %i %s: fseek failed", label, __FILE__, __LINE__, ACF_PATH);
	  (void) errorPMsg(error);
	} 
      }
    } /* while */
    /* Clear the temporary buffer after each line is read in. */
    (void) bzero(tmpbuf, sizeof(tmpbuf));
  }

  /*
   * 7. If the access control file exists but contains no data, then
   *    free the single allocated entry and return an error.
   */
  if (acfentries == 0) {
    (void) freeAcfList(); 
    return(-1);
  } else {
    (void) free(accesscf);      /* must free last empty (allocated) entry */
  }
  /* 
   * 8. Close the access control file.
   */
  if ((fclose(acf)) != 0) {
    (void) perror(ACF_PATH);
    return(-1);
  }
  prev->next = NULL;
  accesscf = head;

  return(0);
} /* parseAcf */


/*
 * 1. Determine if "user" has permission to execute "command".
 * 2. Remove any command line arguments given for this command,
 *    if acfentry->args = TRUE.
 * 3. Prompt user with a reason for running this commandm
 *    if acfentry->reason = TRUE.
 * 4. Set argv[0] to be the absolute path and set runuid/rungid
 *    if it isn't "*" as listed in the access control file.
 * 5. The absolute filepath cannot be world writable.
 * 6. The absolute path cannot be group writable.
 * 7. The absolute filepath cannot be a link.
 * 8. Free all allocated memory for the access control entry list. 
 *
 * Returns 0 if user is authorized to execute "command" 
 *   Sets runuidptr to the user to run the process as.
 *   Sets rungidptr to the group to run the process as.
 * Returns -1 if user does not have permission to execute "command".
 *
 * int *argc;         Number of command line arguments
 * char *argv[];      Command line arguments
 * char *runuidptr;   User to run the process as
 * char *rungidptr;   Group to run the process as
 * const char *user;  Username requesting to run command 
 */
PUBLIC int checkCommandAccess(int *argc, 
			      char **argv, 
			      char *runuidptr, 
			      char *rungidptr, 
			      const char *user)
{
  accessentry *acfentry = NULL; /* "command" access control entry   */
  size_t path_size;             /* size of absolute pathname buffer */
  size_t path_len;              /* length of NULL terminated path   */
#if (TIGHT_CHECKING == TRUE)
  struct stat st;               /* "command" file attributes */
#endif /* TIGHT_CHECKING */

  /* 
   * 1. Determine if "user" has permission to execute "command".
   */
  if ((acfentry = findAcfEntry(*argv)) == NULL) {
    /* argv[0] does not have an access control file entry. */
    return(-1);
  }
  else {
    if (inAccessList(acfentry->users, user) != 0)
      /* User is not priviledged to run argv[0]. */
      return(-1);
    if ((acfentry->argv) != TRUE) {
      /*
       * 2. Remove any command line arguments given for this command,
       *    if acfentry->args = TRUE.
       */
      (void) removeOptions(argc, argv);
    }
#if (USE_SYSLOG == TRUE)
    if ((acfentry->reason) != FALSE) {
      /*
       * 3. Prompt user with a reason for running this command,
       *    if acfentry->reason = TRUE.
       */
      (void) promptReason();
    }
#endif /* USE_SYSLOG */

    /*
     * 4. Set argv[0] to be the absolute path and set runuid/rungid
     *    if it isn't "*" as listed in the access control file.
     */
    if (strncmp(acfentry->absolute, "*", 1) != 0) {
      argv[0] = __pathalloc(&path_size);
      path_len = path_size - 1;
      (void) strncpy(argv[0], acfentry->absolute, path_len);

#if (TIGHT_CHECKING == TRUE)
      if (lstat(acfentry->absolute, &st) != 0) {
	(void) esprintf(error, "%s: \"%s\", line %i %s lstat failed", label, __FILE__, __LINE__, ACF_PATH);
	(void) errorPMsg(error);
      }
      
      /*
       * 5. The absolute path cannot be a link.
       */
      if (((st.st_mode) & S_IFMT) == S_IFLNK) {
	(void) esprintf(error,"%s is a symbolic link and will not be executed.\n", acfentry->absolute);
	(void) errorMsg(error, FALSE);
      }
      
      /*
       * 6. the absolute path cannot be world writable.
       */
      if (st.st_mode & S_IWOTH) {
	(void) esprintf(error,"%s is world writable and will not be executed.\n", acfentry->absolute);
	(void) errorMsg(error, FALSE);
      }

      /*
       * 7. The absolute path cannot be group writable.
       */
      if (strncmp(acfentry->absolute, "*", 1) != 0) {
        if (st.st_mode & S_IWGRP) {
	  (void) esprintf(error,"%s is group writable and will not be executed.\n", acfentry->absolute);
	  (void) errorMsg(error, FALSE);
	}
      }

#endif /* TIGHT_CHECKING */
    }
    (void) strncpy(runuidptr, acfentry->runuid, MAXUSERNAMELEN);
    (void) strncpy(rungidptr, acfentry->rungid, MAXGROUPNAMELEN);
  }
  /* 
   * 8. Free all allocated memory for the access control entry list. 
   */
  (void) freeAcfList(); 

  /* User is authorized to execute command. */
  return(0);
} /* checkCommandAccess */

/*
 * 1. Locate a user in an access list.
 *
 * Returns 0 if user is found in the access list.
 * Returns -1 if user is not found in the access list.
 *
 * alist *accesslist;    Access list to search.
 * const char *user;     User to locate in an access list. 
 */          
PUBLIC int inAccessList(alist *accesslist, const char *user)
{
  /*
   * 1. Locate a user in an access list.
   */
  while (accesslist != NULL){
    if (strncmp(accesslist->name, user, MAXUSERNAMELEN) == 0)
      return(0);
    accesslist = accesslist->next;
  }
  return(-1);
} /* inAccessList */

/*
 * 1. Locate the acfentry with basename or absolute filename
 *    equal to "command".
 *
 * Returns NULL if an acfentry is not found for "command".
 * Returns an accessentry pointer of the first acfentry containing "command". 
 *
 * const char *command;                Command to locate in an access entry
 */
PRIVATE accessentry *findAcfEntry(const char *command)
{
  accessentry *acfentry = accesscf; /* Beginning of access entry list       */

  /*
   * 1. Locate the acfentry with basename or absolute filename
   *    equal to "command".
   */
  while (acfentry != NULL){
    if ((strncmp(acfentry->mnemonic, command, maxpathlen) == 0) ||
	(strncmp(acfentry->absolute, command, maxpathlen) == 0) ||
	((strncmp(acfentry->mnemonic, "*", 1) == 0) &&
	 (strncmp(acfentry->absolute, "*", 1) == 0)))
      return(acfentry);
    acfentry = acfentry->next;
    }
  return(NULL);

} /* findAcfEntry */

/*
 * 1. Find user in the access list.
 * 2. Remove user from the access list.
 *
 * Returns 0 if the user was deleted from the access list.
 * Returns -1 if user not found in the access list.
 *
 * const char *user;                 User to remove from the access list.
 * accessentry *acfentry;            Access control entry to modify.
 */
PRIVATE int removeFromAccessList(const char *user, 
				 accessentry *acfentry)
{
  alist *accesslist = acfentry->users; /* Current access list element      */
  alist *prevaccesslist = NULL;        /* Previous access list element     */

  /* 
   * 1. Find user in the access list.
   */
  while ((accesslist->name != NULL) && 
         (strncmp(accesslist->name, user, MAXUSERNAMELEN) != 0)) {
    prevaccesslist = accesslist;
    accesslist = accesslist->next;
  }

  /* User not found in access list. */
  if (accesslist == NULL) 
    return(-1);
  else {
    /*
     * 2. Remove user from the access list. 
     */
    if (prevaccesslist == NULL) {
      /* Remove user from front of the list. */
      acfentry->users = acfentry->users->next;
    }
    else if (accesslist->next == NULL) {
      /* Remove user from end of the list. */
      prevaccesslist->next = NULL;
    }
    else {
      /* Remove user from middle of the list. */
      prevaccesslist->next = accesslist->next;
    }
    (void) free(accesslist);
    return(0); 
  }
} /* removeFromAccessList */

/*
 * 1. Verify user exists in the passwd file.
 * 2. Find the end of the access list.
 * 3. Add user to the access list.
 *
 * const char *user;       User to add to the access list.
 * accessentry *acfentry;  Access control entry to modify.
 */
PRIVATE void addToAccessList(const char *user, accessentry *acfentry)
{
  alist *accesslist = acfentry->users; /* Current access list element  */
  alist *prevaccesslist = NULL;        /* Previous access list element */
  char newuser[MAXUSERNAMESIZE];      /* Verified user name           */

  /*
   * 1. Verify user exists in the passwd file.
   */
  (void) bzero(newuser, MAXUSERNAMESIZE);
  (void) strncpy(newuser, user, MAXUSERNAMELEN);
  (void) validateUser(newuser);

  /* 
   * 2. Find the end of the access list.
   */
  if (accesslist != NULL) {
    while (accesslist != NULL) {
      prevaccesslist = accesslist;
      accesslist = accesslist->next;
    }
    accesslist = prevaccesslist-> next = initAcfUserEntry();
  }
  else {
    /* First entry in the list. */
    accesslist = acfentry->users = initAcfUserEntry();
  }
  /*
   * 3. Add user to the access list.
   */
  (void) strncpy(accesslist->name, newuser, MAXUSERNAMELEN); 
} /* addToAccessList */

/*
 * 1. Create an access user list entry.
 * 2. Initialize an access user list entry.
 * 3. Return the access user list.
 */
PRIVATE alist *initAcfUserEntry(void)
{
  alist *alistentry; /* access control file access list */

  /* 
   * 1. Create an access user list entry.
   */
  alistentry = (alist *) __ecmalloc(sizeof(alist));

  /* 
   * 2. Initialize an access user list entry.
   */
  (void) bzero(alistentry->name, MAXUSERNAMESIZE);
  /* Initial user should not be a real user. */
  (void) strncpy(alistentry->name, DEVNULL, MAXUSERNAMELEN);
  alistentry->next = NULL;

  /*
   * 3. Return the access user list.
   */
  return(alistentry);
} /* initAcfUserEntry */

/*
 * 1. Allocate a single access control file entry.
 * 2. Initialize access control file entry. 
 *
 * uid_t defaultuid;                  Initial user  id
 * gid_t defaultgid;                  Initial group id                   
 */ 
PRIVATE accessentry *initAcfEntry(uid_t defaultuid, gid_t defaultgid)
{
  accessentry *acfentry;           /* A single access control file entry */
  size_t path_size;                /* size of absolute pathname buffer */

  /*
   * 1. Allocate a single access control file entry.
   */
  acfentry = (accessentry *) __ecmalloc(sizeof(accessentry));

  /* 
   * 2. Initialize access control file entry. 
   */
  acfentry->mnemonic = __pathalloc(&path_size);
  acfentry->absolute = __pathalloc(&path_size);
  acfentry->reason = TRUE; /* ask for a reason by default */ 
  acfentry->argv = FALSE;  /* no parameters allowed to be passed by default */
  (void) bzero(acfentry->runuid, sizeof(acfentry->runuid));
  (void) bzero(acfentry->rungid, sizeof(acfentry->rungid));
  (void) snprintf(acfentry->runuid, MAXUSERNAMELEN, "%lu", (long unsigned)defaultuid);
  (void) snprintf(acfentry->rungid, MAXGROUPNAMELEN, "%lu", (long unsigned)defaultgid);
  acfentry->users = NULL;  /* list of users with access for this command */
  acfentry->next = NULL;   /* the next acf entry */

  return(acfentry);
} /* initAcfEntry */

/*
 * 1. Free all memory associated with the access control file entries.
 */
PUBLIC void freeAcfList(void)
{
  accessentry *acfentry; /* First access control file entry */

  /*
   * 1. Free all memory associated with the access control file entries.
   */
  while (accesscf != NULL) {
    acfentry = accesscf;
    accesscf = accesscf->next;
    (void) free(acfentry);
  }
} /* freeAcfList */

#if (TIGHT_CHECKING == TRUE)
/*
 * Checks various attributes of the access control file.
 *
 * 1. The access control file must exist.
 * 2. The access control file cannot be a relative path.
 * 3. The access control file cannot be a link.
 * 4. The access control file must be owned by ACF_UID.
 * 5. The access control file must be owned by ACF_GID.
 * 6. The access control file must have a mode of ACF_MODE.
 */
PUBLIC void security(void)
{
  struct stat st; /* access control file attributes */

  /*
   * 1. The access control file must exist.
   */
  if (lstat(ACF_PATH, &st) != 0) {
    (void) esprintf(error, "%s: \"%s\", line %i %s lstat failed", label, __FILE__, __LINE__, ACF_PATH);
    (void) errorPMsg(error);
  }

  /*
   * 2. The access control file cannot be a relative path.
   */
  if (strncmp(ACF_PATH, "/", 1) != 0) {
    (void) esprintf(error, "%s is a relative pathname and will not be executed.", ACF_PATH);
    (void) errorMsg(error, FALSE);
  }

  /*
   * 3. The access control file cannot be a link.
   */
  if (((st.st_mode) & S_IFMT) == S_IFLNK) {
    (void) esprintf(error, "%s is a symbolic link and will not be executed.", ACF_PATH);
    (void) errorMsg(error, FALSE);
  }

  /* 
   * 4. The access control file must be owned by ACF_UID.
   */
  if (ACF_UID != st.st_uid) {
    (void) esprintf(error, "The uid of %s is not %lu.", ACF_PATH, (long unsigned)ACF_UID);
    (void) errorMsg(error, FALSE);
  }

  /*
   * 5. The access control file must be owned by ACF_GID.
   */
  if (ACF_GID != st.st_gid) {
    (void) esprintf(error, "The gid of %s is not %lu.", ACF_PATH, (long unsigned)ACF_GID);
    (void) errorMsg(error, FALSE);
  }
  
  /* 
   * 6. The access control file must have a mode of ACF_MODE.
   */
  if ((ACF_MODE) != st.st_mode) {
    (void) esprintf(error,"The file mode of %s is not %o.", ACF_PATH, ACF_MODE);
    (void) errorMsg(error, FALSE);
  }
} /* security */
#endif /* TIGHT_CHECKING */

/*
 * 1. Display mnemonic and absolute pathname.
 * 2. Display users that can run this command, only if the user is not
 *    dependant on the access control file for access.
 * 3. User and group to run the process as.
 * 4. Display if command line arguments are passed to the command. 
 * 5. Display if the user needs to specifiy a reason to run this command.
 *
 * accessentry *entry;          access control entry to display
 * BOOLEAN needACF;             is the access control file needed for access? 
 */
PUBLIC void printAcfCommand(accessentry *entry, BOOLEAN needACF)
{
  alist *accesslist;        /* access control file user access list          */

  /* 
   * 1. Display mnemonic and absolute pathname.
   */
  (void) printf("%s -> %s\n", entry->mnemonic, entry->absolute);

  /* 
   * 2. Display users that can run this command, only if the user is not
   *    dependant on the access control file for access.
   */
  if (needACF == FALSE) {
    (void) printf("  Authorized users:");
    accesslist = entry->users;
    while (accesslist != NULL) {
      (void) printf(" %s", accesslist->name);
      accesslist = accesslist->next;
    }
    (void) printf("\n");
  }

  /* 
   * 3. User and group to run the process as.
   */ 
  (void) printf("  Process id: user=%s, group=%s\n", entry->runuid, entry->rungid);

  /* 
   * 4. Display if command line arguments are passed to the command. 
   */
  if (entry->argv == TRUE) {
    (void) printf("  Arguments allowed.");
  } else {
    (void) printf("  No arguments allowed.");
  }

  /* 
   * 5. Display if the user needs to specifiy a reason to run this command.
   */
  if (entry->reason == TRUE) {
    (void) printf("  Reason for running command must be specified.\n");
  } else {
    (void) printf("  No reason is needed.\n");
  }
} /* printAcfCommand */ 

#endif /* ifdef USE_ACF acf.c */

/* end acf.c */
