/* config.h -- compile-time administrator-set configuration options.
 *
 * $Id: config.h,v 1.26 1998/11/12 22:38:45 ccarp Exp $
 * 
 * runas - execute a single command as a different user and/or group
 * Copyright (C) 1998 Chris Carpinello
 *
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 * Send all bugs, changes, and problems to the maintainer listed in
 * "runas --info" output.
 */

/* 
 * This is the configuration file for runas.  Any in-depth explanation
 * of each of the compile-time configuration options is detailed in the
 * GNU texinfo file runas.info.
 *
 * Some notes about this file:
 *
 * Defines with boolean settings in this configuration file can be set 
 * to TRUE or FALSE.  Defines marked as "must be defined" must have
 * a valid value.  Defines marked as "define is optional" do not
 * require that the symbolic constant be defined.
 */

/* 
 * Many sites have a group that all local system administrators are a
 * member of.  For a more secure environment, it is advisable that all
 * local system administrators that belong to this common administrative
 * group have this group as their base group, which is defined in their
 * passwd entry.  Runas takes precautions by not allowing access to anyone 
 * who does not have have SYSADMIN_GID as their base gid (thus, sgid 
 * SYSADMIN_GID shells can't spoof runas).  Additionally, anyone who is a 
 * member of group SYSADMIN_GID does not rely on the existance of the access 
 * control file (if enabled) so if the access file were to be removed or 
 * corrupted, administrative functions could still be performed. 
 *
 * Any user that has a base group of SYSADMIN_GID can run any program as
 * any user on the system.
 *
 * ALLOW_GROUPIES can be set to TRUE to allow users who do not have 
 * SYSADMIN_GID as their base gid (but have SYSADMIN_GID as a supplemental
 * group) to use runas.
 */
#define	SYSADMIN_GID        20      /* must be defined */
#define ALLOW_GROUPIES      FALSE   /* must be defined */

/*
 * At sites where multiple administrators with various levels of privlidges
 * are performing system administrative functions, it's often useful if all 
 * files newly created by root are owned by a system administration group.
 * If ROOT_GID is set, the gid specified to will be used instead of root's 
 * base gid when creating files.  Specifying "runas -root" will override
 * this value if it is set.  By default, set the ROOT_GID to the SYSADMIN_GID.
 */
#define ROOT_GID            SYSADMIN_GID  /* define is optional */

/*
 * Define VALIDATE_USER_GROUP to TRUE to only allow id changes to valid
 * users and groups as defined in the passwd and group files.  Setting
 * this value to FALSE will allow processes to be run as any uid and gid,
 * even if the uid or gid does not have a corresponding entry in the
 * passwd or group files.
 */
#define VALIDATE_USER_GROUP TRUE         /* must be defined */

/*
 * Ok_Folks contains a comma separated list of uids that bypass syslog 
 * when runas is invoked.  This was originally introduced to only include
 * 0 (root) since it's nearly pointless to syslog actions that root is
 * performing (when runas is used as habit while in a root shell).   
 */
#undef OK_FOLKS                          /* define is optional */

/*
 * Define USE_UID_RANGE to TRUE to establish a range of uids that are 
 * authorized to use runas in addition to the SYSADMIN_GID restriction.
 *
 * If USE_UID_RANGE is defined, then only those users with uids
 * between and including UID_RANGE_BOTTOM and UID_RANGE_TOP will bypass 
 * the access control file (must be valid uid values).  Additionally,
 * all uids in the range must have a base gid (the user's gid in the passwd 
 * file) of SYSADMIN_GID, even if ALLOW_GROUPIES is defined to TRUE.
 */
#define USE_UID_RANGE          FALSE      /* must be defined */
#if (USE_UID_RANGE == TRUE)        
#define UID_RANGE_BOTTOM       0          /* lowest uid allowed  */
#define UID_RANGE_TOP          0          /* highest uid allowed */
#endif

/*
 * PRESERVE_ENVIRONMENT_DEFAULT defines the default behavior when a 
 * restricted shell (one listed in /etc/shells) is specified.
 *
 * Setting to TRUE will preserve the shell environment (all environmental
 * variables) when a shell is run.  
 *
 * Setting to FALSE will remove all shell environmental variables except
 * for TERM, HOME, SHELL, SHLVL, USER, LOGNAME and PATH when a shell is run.
 *
 * Note: The default behavior can be overridden via the command line 
 * arguments --preserve and --sanitize.
 */
#define PRESERVE_ENVIRONMENT_DEFAULT  TRUE  /* must be defined */

/*
 * Define USE_SECURE_PATH to TRUE to prepend a "secure" system path
 * to an administrators's existing path.  It is strongly recommended
 * that this feature is enabled and meaningfully defined.
 *
 * SECURE_PATH is used as the default $PATH when a shell environment
 * is sanitized (--sanitize, --login) so it is important that SECURE_PATH
 * be meaningfully defined.
 *
 * The SECURE_PATH defined here is a sample suggested secure path.
 * Change this to fit your system.  
 */
#define USE_SECURE_PATH TRUE                /* must be defined */
#if (USE_SECURE_PATH == TRUE)
#define SECURE_PATH "/usr/local/sbin:/sbin:/usr/sbin:/usr/local/bin:/bin:/usr/bin" /* do not need trailing : */
#endif

/* 
 * Define IGNORE_PREFIX_IN_PATH as a colon separated list of directory
 * prefixes that will be ignored when determining the absolute path
 * of the command to run (no trailing colon is needed).  All directories
 * beneath and including the listed directories will be ignored.
 *
 * Example: If IGNORE_PREFIX_IN_PATH="/tmp:/usr/tmp", PATH=/tmp/bin:/usr/bin,
 * and /tmp/bin/sh and /usr/bin/sh exist, then /tmp/bin/sh will be ignored 
 * and /usr/bin/sh will be run instead.  However, if /tmp/bin/sh is explicitly 
 * given as the command to execute, then /tmp/bin/sh will be run.
 */
#define IGNORE_PREFIX_IN_PATH ".:/tmp:/usr/tmp:/var/tmp:/home"

/*
 * Define IGNORE_SYMLINKS to TRUE to ignore any files in $PATH that are
 * symbolic links.
 *
 * Warning: some packages use symbolic links to cause binaries to behave
 * differently based on the filename they are invoked with.
 */
#define IGNORE_SYMLINKS FALSE               /* must be defined */ 

/*
 * Define EXEC_ONLY_ADMIN_OWNED_FILES to TRUE to only execute files that are
 * owned by one of the usernames listed in EXEC_ONLY_FILES_OWNED_BY.
 * Files owned by another username will be skipped (if $PATH is being
 * searched) and will not be run even if the filepath is explicitly defined.
 * Additionally, any $PATH directory being searched in an attempt to fully
 * qualify the path for "command" will be skipped if the directory is not
 * owned by a username in EXEC_ONLY_FILES_OWNED_BY.
 */
#define EXEC_ONLY_ADMIN_OWNED_FILES TRUE    /* must be defined */
#if (EXEC_ONLY_ADMIN_OWNED_FILES == TRUE)
#define EXEC_ONLY_FILES_OWNED_BY "root", "bin" /* quoted, comma separated */
#endif /* EXEC_ONLY_ADMIN_OWNED_FILES */

/*
 * Define ALLOW_ROOT_SHELLS to TRUE to allow an administrator to execute
 * a root shell.  If set to FALSE, then root shells cannot be executed by 
 * the administrator.
 *
 * Note: Untrustworthy administrators will be able to circumvent this feature 
 * by running a copy of a shell from a path not listed in /etc/shells.
 * Read the info file for complete details.
 *
 * Note: While this feature examines /etc/shells, some systems
 * may have /usr/bin symlinked to /bin (or vice versa).  If this
 * is the case on your system, make sure that both /bin and /usr/bin
 * entries exist for each shell in /etc/shells.
 */
#define ALLOW_ROOT_SHELLS TRUE             /* must be defined */

/*
 * Define USE_ACF to TRUE to enable runas to perform ``sudo-like'' access 
 * control restrictions via an access control file (ACF).  This allows 
 * restriction of commands on a command by command basis.  See 
 * "runas.info" for a complete description of the access control 
 * file.
 */
#define USE_ACF FALSE                     /* must be defined */

/*
 * If USE_ACF and TIGHT_CHECKING are set to TRUE, certain attributes of the 
 * ACF (ownership, permissions, can't be a link) and the absolute pathname 
 * (in the ACF) are verified before executing anything.  See "runas.info"
 * for a complete description of the access control file (ACF).
 * 
 * ACF_MODE is mode of the installed ACF file.  Optimum secure permission 
 *   mode is 0400. see macros listed in stat(2) to change the mode
 * ACF_UID is the user owner (uid) of the ACF 
 * ACF_GID is the group owner (gid) of the ACF
 */
#if (USE_ACF == TRUE)
#define TIGHT_CHECKING TRUE
#define ACF_MODE     S_IFREG|S_IRUSR           /* file mode 0400 */
#define ACF_UID      DEFAULT_UID               /* root user */
#define ACF_GID      SYSADMIN_GID              /* system administrator group */
#endif /* USE_ACF */

/* end config.h */
