/* log.c -- syslog logging routines
 *
 * $Id: log.c,v 1.8 1998/10/16 22:05:39 ccarp Exp $
 * 
 * runas - execute a single command as a different user and/or group
 * Copyright (C) 1998 Chris Carpinello
 *
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 * Send all bugs, changes, and problems to the maintainer listed in
 * "runas --info" output.
 */

/* System headers */
#include <stdio.h>
#if (HAVE_STDLIB_H == 1)
#include <stdlib.h>
#endif /* HAVE_STDLIB_H */
#include <pwd.h>
#include <grp.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <ctype.h>
#if (HAVE_LIMITS_H == 1)
#include <limits.h>
#endif /* HAVE_LIMITS_H */
#if (USE_SYSLOG == 1)
#include <sys/utsname.h>
#if (HAVE_SYSLOG_H == 1)
#include <syslog.h>
#else 
#include <sys/syslog.h>
#endif /* HAVE_SYSLOG_H */
#endif /* USE_SYSLOG */
#include <string.h>
#include <errno.h>
#if (HAVE_UNISTD_H == 1)
#include <unistd.h>
#endif /* HAVE_UNISTD_H */

/* Local headers and macros */
#include "defines.h"
#include "config.h"
#include "acf.h"
#include "log.h"

/* ifdef log.c */
#if (USE_SYSLOG == TRUE)  /* Nothing to compile unless this is set */

/* File scope variables */
char *tty        = NULL;     /* Current tty in use                 */
char *ttymalloc  = NULL;     /* Malloced space for *tty            */
char *syslog_buf = NULL;     /* Syslog format buffer               */
size_t syslog_bufsize;       /* Size of syslog format buffer       */
size_t syslog_labelen;       /* Length of syslog() identity label  */
char default_syslog_label[] = "runas"; /* Default syslog label     */

/* External variables */
#include "extern.h"

/* Local headers and macros */
#include "proto.h"

/* Functions */

/*
 * 1. Surround each command line argument with braces to better distinguish
 *    multi-word quoted command line arguments from whitespace delimited
 *    arguments.
 *
 * It is implied that "msg" is large enough to contain each argument
 * plus the 3 additional characters per argument.  This function performs 
 * no boundary checking.
 *
 * char *msg;           String to hold delimited arguments
 * char * const *args;  Arguments to delimit
 */
PUBLIC void delimitArgs (char *msg, char * const *args)
{
  /*
   * 1. Surround each command line argument with braces to better distinguish
   *    multi-word quoted command line arguments from whitespace delimited
   *    arguments.
   */
  while (*args != NULL) {
    (void) strncat(msg, " [", 2);
    (void) strcat(msg, *args);
    (void) strncat(msg, "]", 1);
    args++;
  }
} /* delimitArgs */

/*
 * 1. Check boundary length of executable name.
 * 2. Initialize syslog logging.
 * 3. Hostname string length needed to determine syslog buffer size.
 * 4. Dynamically allocate memory for maximum, fixed-length syslog format 
 *    buffer based on the system's maximum syslogd message (SYSLOG_BUFSIZE), 
 * 5. Determine the tty in use.
 *
 * Global variables set: 
 *   hostnamelen     -- string length of hostname
 *   syslog_labelen  -- string length of syslog label
 *   syslog_bufsize  -- string length of digestable syslog buffer
 *   syslog_buf      -- digestable syslog buffer
 *
 * const char *label;        Syslog label       
 */
PUBLIC void initSyslog(const char *label)
{
  struct utsname *host;  /* uname information  */
  size_t hostnamelen;    /* strlen of hostname */
  size_t nullttylen;     /* strlen of NULLTTY  */

  /* 
   * 1. Check boundary length of executable name.
   */
  if (strlen(label) > SYSLOG_LABEL_MAX) {
    free(label);
    label = default_syslog_label;
    (void) fprintf(stderr, "%s: argv[0] name length is longer than maximum path length allowed.  Using ``%s'' instead.\n", label, label);
  }

  /* 
   * 2. Initialize syslog logging.
   */
  syslog_labelen = strlen(label);
  (void) openlog(label, SYSLOG_OPTIONS, SYSLOG_FACILITY);

  /* 
   * 3. Hostname string length needed to determine syslog buffer size.
   */
  host = (struct utsname*) __ecmalloc(sizeof(struct utsname));
  if (uname(host) <= -1) { /* Solaris returns non-negative for success */
    (void) esprintf(error, "%s: \"%s\", line %i uname failed", label, __FILE__, __LINE__);
    (void) errorPMsg(error);
  } 
  /* Note: nodename may be fully qualified. */
  hostnamelen = strlen(host->nodename);
  (void) free(host);

  /* 
   * 4. Dynamically allocate memory for maximum, fixed-length syslog format 
   *    buffer based on the system's maximum syslogd message (SYSLOG_BUFSIZE), 
   *
   * Take into account some "fixed" lengths like time, date, and separaters 
   * (SYSLOG_FIXEDLEN), hostname (hostnamelen), and the syslog identity 
   * label (syslog_labelen).
   */
  syslog_bufsize = SYSLOG_BUFSIZE - SYSLOG_FIXEDLEN - hostnamelen - syslog_labelen;
  if (syslog_bufsize <= (size_t)0) {
    (void) errorMsg("Syslog buffer length too small.", FALSE);
  }
  syslog_buf = __stralloc(syslog_bufsize);

  /* 
   * 5. Determine the tty in use.
   */
  tty = ttyname(fileno(stdin));
  /* Note: ttyname returns NULL through a pipe or remote shell usage. */
  /* Note: SYSV printf() SIGSEGVs on NULL pointers.                   */
  if (tty == NULL) {   
    nullttylen = strlen(NULLTTY);
    ttymalloc = __stralloc(nullttylen + 1); /* +1 for NULL */ 
    (void) strncpy(ttymalloc, NULLTTY, nullttylen);
    tty = ttymalloc;
  }
} /* initSyslog */

/*
 * 1. Initialize reason value to NULL.
 * 2. Prompt user for reason to run command.
 *
 * Global variable set
 *   reason  -- reason user is running this command.
 */
#if (USE_ACF == TRUE)
PUBLIC void promptReason(void)
{
  size_t len = 0; /* strlen of reason given to run command */

  /* 
   * 1. Initialize reason value to NULL.
   */
  (void) bzero(reason, MAXREASONSIZE);

  /*
   * 2. Prompt user for reason to run command.
   */
  while (len < 4) {  /* so they can't just hit return when giving a reason */
    (void) printf("Reason: ");
    (void) fflush(stdout);
    if (fgets(reason, MAXREASONLEN, stdin) == NULL) {
      (void) errorMsg("fgets() failure in promptReason().", FALSE);
    }
    len = strlen(reason);
  }
} /* promptReason */
#endif /* USE_ACF */

/*
 * Tie up loose ends with syslog and syslog variables.
 *
 * 1. Close the syslog() connection. 
 * 2. Free dynamically allocated buffer for syslog format parameter.
 * 3. Free dynamically allocated buffer for "(null)" ttyname 
 */
PUBLIC void syslogCleanUp(void)
{
  /* 
   * 1. Close the syslog() connection. 
   */
  (void) closelog();

  /* 
   * 2. Free dynamically allocated buffer for syslog format parameter.
   */
  if (syslog_buf != NULL) {
    (void) free(syslog_buf);
  }

  /* 
   * 3. Free dynamically allocated buffer for "(null)" ttyname 
   */ 
  if (ttymalloc != NULL) {
    (void) free(ttymalloc);
  }
} /* syslogCleanup */

/*
 *  1. Do not syslog commands of users listed in Ok_Folks.
 *  2. Save the user name of the user executing runas.
 *  3. Save the user name of the user to run the process as.
 *  4. Save the group name of the group to run the process as.
 *  5. Determine the working directory.
 *  6. Determine how much memory to allocate for command line options.
 *  7. Allocate memory for syslog message.
 *  8. Format the syslog message into the message buffer.
 *  9. Surround each command line argument with braces to better 
 *     distinguish multi-word quoted command line arguments from 
 *     whitespace delimited arguments. 
 * 10. Syslog the message.
 * 11. Free dynamically allocated buffers.
 *
 * uid_t myuid;
 * uid_t myeuid;
 * uid_t runuid;
 * gid_t rungid;
 * int argc;                          number of command line arguments
 * char * const *argv;                command line arguments               
 */
PUBLIC void syslogCommand(uid_t myuid, 
			  uid_t myeuid, 
			  uid_t runuid, 
			  gid_t rungid, 
			  int argc, 
			  char * const *argv)
{
  struct passwd *pw;              /* passwd structure                     */
  struct group *group;            /* group structure                      */
  int i;                          /* iterate through Ok_Folks             */
  int msglen;                     /* sizeof larger SYSLOG_MSG macro       */
  size_t argvlen;                 /* size_t of all command line arguments */
  char fromuser[MAXUSERNAMESIZE]; /* username of user invoking runas      */
  char touser[MAXUSERNAMESIZE];   /* username of new process              */
  char togroup[MAXGROUPNAMESIZE]; /* groupname of new process             */
  char *cwd;                      /* current working directory            */ 
  char *syslog_msg;               /* formatted syslog message with args   */ 
  size_t syslog_msgsize;          /* size_t of syslog message buffer      */ 

  /* 
   *  1. Do not syslog commands of users listed in Ok_Folks.
   */
  for (i=0; (Ok_Folks[i] != OK_FOLKS_DELIMITER) || (myuid == Ok_Folks[i]); i++); 
  if (Ok_Folks[i] != myuid) { 

    /* 
     *  2. Save the user name of the user executing runas.
     */
    pw = getpwuid(myuid);
    if (pw == NULL) { /* ENOMEM */
      (void) esprintf(error, "%s: \"%s\", line %i getpwuid failed", label, __FILE__, __LINE__);
      (void) errorPMsg(error);
    }
    (void) bzero(fromuser, MAXUSERNAMESIZE);
    (void) strncpy(fromuser, pw->pw_name, MAXUSERNAMELEN);
    
    /* 
     *  3. Save the user name of the user to run the process as.
     */
    (void) bzero(touser, MAXUSERNAMESIZE);
    pw = getpwuid(runuid);
    if (pw == NULL) { /* No such user or ENOMEM */
#if (VALIDATE_USER_GROUP == TRUE)
      if (errno == ENOMEM) {
	(void) esprintf(error, "%s: \"%s\", line %i getpwuid failed", label, __FILE__, __LINE__);
	(void) errorPMsg(error);
      }
#else
      (void) snprintf(touser, MAXUSERNAMELEN, "%lu", (long unsigned)runuid);
#endif /* VALIDATE_USER_GROUP */
    } else {
      (void) strncpy(touser, pw->pw_name, MAXUSERNAMELEN);
    }
    
    /* 
     *  4. Save the group name of the group to run the process as.
     */
    (void) bzero(togroup, MAXGROUPNAMESIZE);
    group = getgrgid(rungid);
    if (group == NULL) { /* No such group or ENOMEM */
#if (VALIDATE_USER_GROUP == TRUE)
      if (error == ENOMEM) {
	(void) esprintf(error, "%s: \"%s\", line %i getgrgid failed", label, __FILE__, __LINE__);
	(void) errorPMsg(error); 
      } 
#else
      (void) snprintf(togroup, MAXGROUPNAMELEN, "%lu", (long unsigned)rungid);
#endif /* VALIDATE_USER_GROUP */
    } else {
      (void) strncpy(togroup, group->gr_name, MAXGROUPNAMELEN);
    }
    
    /* 
     *  5. Determine the working directory.
     */
    cwd = getCurrentWorkingDirectory(myuid, myeuid);
    
    /* 
     *  6. Determine how much memory to allocate for command line options.
     */
    argvlen = 0;
    for (i=0; i < argc; i++) {
      /* Add 3 for "[", "]" and " " for each argument */
      argvlen += strlen(argv[i]) + 3; 
    }
    /* Allocate storage for larger of two messages */
    msglen = (sizeof(SYSLOG_MSG) > sizeof(SYSLOG_REASONMSG)) ? sizeof(SYSLOG_MSG) : sizeof(SYSLOG_REASONMSG);
    /* argvlen = size_t of all command line arguments. 
     * strlen(tty, cwd, from, user, group, reason) + 1 for NULL. */
    syslog_msgsize = msglen + strlen(tty) + strlen(cwd) + strlen(fromuser) + strlen(touser) + strlen(togroup) + argvlen + 1;
#if (USE_ACF == TRUE)
    if (reason != NULL) {
      syslog_msgsize += strlen(reason);
    }
#endif /* USE_ACF */

    /* 
     *  7. Allocate memory for syslog message.
     */
    syslog_msg = __stralloc(syslog_msgsize); 

    /* 
     *  8. Format the syslog message into the message buffer.
     */
#if (USE_ACF == TRUE)
    if (strlen(reason))
      (void) sprintf(syslog_msg, SYSLOG_REASONMSG, SYSLOG_REASONMSG_ARGS);
    else
#endif /* USE_ACF */
      (void) sprintf(syslog_msg, SYSLOG_MSG, SYSLOG_MSG_ARGS);

    /* 
     *  9. Surround each command line argument with braces to better 
     *     distinguish multi-word quoted command line arguments from 
     *     whitespace delimited arguments. 
     */
     (void) delimitArgs(syslog_msg, argv);

     /* 
      * 10. Syslog the message.
      */
     (void) syslogMsg(SYSLOG_PRIORITY_OKPRIV, syslog_msg);

     /* 
      * 11. Free dynamically allocated buffers.
      */
     (void) free(syslog_msg);
     (void) free(cwd);
  } /* if Ok_Folks != myuid */
} /* syslogCommand */

PUBLIC void syslogCommandNOPRIV(int *argc, char **argv, uid_t myuid, uid_t myeuid, char *user) {
  char *cwd = NULL;            /* current working directory            */
  int i;                       /* iterate through argv                 */
  size_t argvlen;              /* size_t of all command line arguments */
  char *syslog_msg = NULL;     /* formatted syslog message with args   */ 
  size_t syslog_noprivmsglen;  /* length of the no priv part of the msg */
  size_t syslog_msgsize;       /* total size of the message buffer */ 
  size_t syslog_msglen;        /* syslog_msgsize - 1 */ 
  struct passwd *pw;           /* user password info */

  if (user == NULL) {
    pw = getpwuid(myuid);
    user = pw->pw_name;
  }

  /* Get the current working directory. */
  cwd = getCurrentWorkingDirectory(myuid, myeuid);

  /* Determine how much memory to allocate for command line options */
  argvlen = 0;
  for (i=0; i < (*argc); i++) {
    /* Add 3 for "[", "]" and " " for each argument */
    argvlen += strlen(argv[i]) + 3; 
  }
  /*  
   * Allocate enough memory for syslog msg.
   * argvlen = size_t of all command line arguments.
   * strlen(SYSLOG_NOPRIVMSG, tty, cwd, user) + 1 for NULL.
   */
  syslog_noprivmsglen = strlen(SYSLOG_NOPRIVMSG) + strlen(tty) + strlen(cwd) + strlen(user) + 1;
  syslog_msgsize = syslog_noprivmsglen + argvlen + 1;
  syslog_msg = __stralloc(syslog_msgsize);
  syslog_msglen = syslog_msgsize - 1;
  (void) snprintf(syslog_msg, syslog_noprivmsglen, SYSLOG_NOPRIVMSG, SYSLOG_NOPRIVMSG_ARGS);
  (void) free(cwd);
  (void) delimitArgs(syslog_msg, argv);
  (void) errorMsg(syslog_msg, TRUE);
}


/* 
 * Wrapper function for syslog() that prevents log spoofing and buffers 
 * output to a length of SYSLOG_BUFSIZE to prevent buffer overflows.
 *
 * 1. Break up a large syslog message into smaller message chunks that 
 *    syslogd can digest if the message is greater than syslog_buflen,
 *    preferably breaking up the message on a space boundary.
 *
 * Globals refernced: 
 *   syslog_labelen  -- string length of syslog label
 *   syslog_buflen   -- string length of digestable syslog buffer
 *   syslog_buf      -- digestable syslog buffer
 */

PUBLIC void syslogMsg(int priority, const char *msg)
{
  char   *delimiter;           /* strrchr pointer to character delimiter */ 
  size_t msglen;               /* syslog parsing stuff */
  int    msg_position;         /* position of last character in syslog_buf */
  size_t nonterminated_syslog_buflen; /* string length of non-NULL terminated
                                         syslog_buf */
 /*
  * 1. Break up a large syslog message into smaller message chunks that 
  *    syslogd can digest if the message is greater than syslog_buflen,
  *    preferably breaking up the message on a space boundary.
  *
  * only deal with syslog_buflen-1 bytes at a time. 
  * copy syslog_buf-1 bytes of msg to syslog_buf. 
  * working backwards, replace the first space occurance with a NULL.
  * increment pointer of msg memory address to msg+[element-1]*sizeof(char)
  * if no space exists, print syslog_buflen-1 characters and incremenet 
  *   pointer of msg memory address to msg+syslog_buflen-1.
  * repeat until msg buffer is empty.
  */

  msglen = strlen(msg);
  nonterminated_syslog_buflen = syslog_bufsize - 1; /* Need 1 for NULL */
  while (msglen > nonterminated_syslog_buflen) {
    (void) strncpy(syslog_buf, msg, nonterminated_syslog_buflen);
    syslog_buf[syslog_bufsize] = (char) NULL;
    delimiter = strrchr(syslog_buf, SPACE);     /* split on space boundary */

    if (delimiter == NULL) {                    /* No space found */
      msg_position = nonterminated_syslog_buflen;
      syslog_buf[msg_position + 1] = (char) NULL;
    } else {
      msg_position = nonterminated_syslog_buflen - strlen(delimiter);
      syslog_buf[msg_position] = (char) NULL;
    }
    (void) syslog(priority, syslog_buf); 
    msglen -= msg_position;
    msg += msg_position + 1;
  }
  if (msg != NULL) {
    (void) syslog(priority, msg); 
  }
} /* syslogMsg */

#endif /* ifdef USE_SYSLOG log.c */
/* end log.c */

