/* misc.c -- miscellanous system-like functions. 
 *
 * $Id: misc.c,v 1.24 1998/11/12 21:54:46 ccarp Exp $
 * 
 * runas - execute a single command as a different user and/or group
 * Copyright (C) 1998 Chris Carpinello
 *
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 * Send all bugs, changes, and problems to the maintainer listed in
 * "runas --info" output.
 *
 */

/* System headers */
#include <stdio.h>
#include <stdarg.h>
#if (HAVE_STDLIB_H == 1)
#include <stdlib.h>
#endif /* HAVE_STDLIB_H */
#include <pwd.h>
#include <grp.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <ctype.h>
#if (HAVE_LIMITS_H == 1)
#include <limits.h>
#endif /* HAVE_LIMITS_H */
#if (USE_SYSLOG == 1)
#if (HAVE_SYSLOG_H == 1)
#include <syslog.h>
#else 
#include <sys/syslog.h>
#endif /* HAVE_SYSLOG_H */
#endif /* USE_SYSLOG */
#if (HAVE_MALLOC_H == 1)
#include <malloc.h>
#endif /* HAVE_MALLOC_H */
#include <string.h>
#include <errno.h>
#if (HAVE_UNISTD_H == 1)
#include <unistd.h>
#endif /* HAVE_UNISTD_H */

/* Local headers and macros */
#include "defines.h"
#include "config.h"
#include "acf.h"
#include "log.h"

/* External variables */
#include "extern.h"

/* External function prototypes */ 
#include "proto.h"

#if (HAVE_GETUSERSHELL != 1)
FILE *fpShells;
PUBLIC char *shells[] = {
  "/sbin/sh", 
  "/bin/sh", 
  "/usr/bin/sh", 
  "/sbin/csh",
  "/bin/csh",
  "/usr/bin/csh",
  "/sbin/ksh",
  "/bin/ksh",
  "/usr/bin/ksh",
  "/sbin/tcsh",
  "/bin/tcsh",
  "/usr/bin/tcsh",
  "/sbin/bash",
  "/bin/bash",
  "/usr/bin/bash",
  NULL
};
PUBLIC char **etcshells = NULL;
#endif /* HAVE_GETUSERSHELL */

#if (EXEC_ONLY_ADMIN_OWNED_FILES == TRUE)
PUBLIC char *ownedby[] = { 
#ifdef EXEC_ONLY_FILES_OWNED_BY
  EXEC_ONLY_FILES_OWNED_BY, 
#endif /* EXEC_ONLY_FILES_OWNED BY */
  NULL 
};
#endif

/* Local function prototypes */ 
PRIVATE BOOLEAN ignorePATH    __P((char *path));
PRIVATE BOOLEAN *execthisfile __P((uid_t owner));

/* Functions */

/*
 * 1. Print the error to stderr.
 * 2. USE_SYSLOG: syslog and free the "no priviledge" message.
 * 3. USE_SYSLOG: Tidy up syslog.
 * 4. Exit with a failure.
 *
 * Note: If freeuhoh is TRUE, then the mem location pointed to uhoh is freed 
 * before exiting.  Overload freeuhoh to only log an error mssage to syslogd
 * when TRUE (there's only one case where we want to syslog errors:
 * when the user has no priviledges. 
 */
PUBLIC void errorMsg(char *uhoh, BOOLEAN freeuhoh)
{
  /*
   * 1. Print the error to stderr.
   */
  (void) fprintf(stderr, "%s: %s\n", label, uhoh); /* print it to standard error */

  /*
   * 2. USE_SYSLOG: syslog and free the "no priviledge" message.
   */
#if (USE_SYSLOG == TRUE)
  if (freeuhoh == TRUE) {
    (void) syslogMsg(SYSLOG_PRIORITY_NOPRIV, uhoh);  /* send it to syslogd */
    (void) free(uhoh); 
  }
  /*
   * 3. USE_SYSLOG: Tidy up syslog.
   */
  (void) syslogCleanUp();

#endif /* USE_SYSLOG */

  /*
   * 4. Exit with a failure.
   */
  (void) exit(EXIT_FAILURE);
} /* errorMsg */

/*
 * 1. USE_SYSLOG: Tidy up syslog.
 * 2. Print system error message to stderr.
 * 3. Exit with the error number value.
 */
PUBLIC void errorPMsg(const char *uhoh)
{
  /*
   * 2. Print system error message to stderr.
   */
  (void) perror(uhoh);

  /*
   * 1. USE_SYSLOG: Tidy up syslog.
   */
#if (USE_SYSLOG == TRUE)
  (void) syslogCleanUp();
#endif /* USE_SYSLOG */

  /*
   * 3. Exit with the error number value.
   */
  (void) exit(errno);
} /* errorPMsg */

/*
 * 1. USE_SYSLOG: Tidy up syslog.
 * 2. Exit with a success.
 */
PUBLIC void exitOK(void)
{
  /*
   * 1. USE_SYSLOG: Tidy up syslog.
   */
#if (USE_SYSLOG == TRUE)
  (void) syslogCleanUp();
#endif /* USE_SYSLOG */

  /*
   * 2. Exit with a success.
   */
  (void) exit(EXIT_SUCCESS);
} /* exitOK */

/*
 * 1. Set environment variable "envar" to value "enval".
 *
 * const char *envar;  Environmental variable
 * const char *enval;  Expected environment variable value 
 */
PUBLIC void setenvar(const char *envar, const char *enval)
{
#if (HAVE_PUTENV == 1)
  static char *string; 
  size_t envar_len = strlen(envar);
  size_t enval_len = strlen(enval);
  size_t totalenv_size = envar_len + enval_len + 2;

  /* 
   * 1. Set the environment variable "envar" to value "enval".
   */
  string = __stralloc(totalenv_size);

  (void) strncpy(string, envar, envar_len);
  (void) strncat(string, "=", 1);
  (void) strncat(string, enval, enval_len);
  if(putenv(string) != 0) {
#else
  if((setenv(envar, enval, OVERWRITE)) != 0) {
#endif /* HAVE_PUTENV */
   (void) esprintf(error, "Unable to set environment variable %s=%s\n", envar, enval);
   (void) errorMsg(error, FALSE); 
  }
} /* setenvar */

/*
 * 1. If user is a uid, verify it is a valid defined uid and then covert 
 *    it to a user name.
 * 2. If user is a user name, then verify it is a valid defined 
 *    user name.
 */
PUBLIC void validateUser(char *user)
{
  struct passwd *pw;
  uid_t uid;

  /*
   * 1. If user is a uid, verify it is a valid defined uid and then covert 
   *    it to a user name.
   */
  if (isanumber(user)) {
    uid = (uid_t)atouid(user);
    pw = getpwuid(uid);
    if(pw == NULL) {   /* no user associated with that uid */
#if (VALIDATE_USER_GROUP == TRUE)      
      (void) esprintf(error,"No such user id `%lu'", (long unsigned)uid);
      (void) errorMsg(error, FALSE);
#endif /* VALIDATE_USER_GROUP */
    }
    else {
      /* Note: assumes user has been allocated enough space already */
      (void) bzero(user, MAXUSERNAMESIZE);
      (void) strncpy(user, pw->pw_name, MAXUSERNAMELEN);
    }
  }
#if (VALIDATE_USER_GROUP == TRUE)      
  else {
    /*
     * 2. If user is a user name, then verify it is a valid defined 
     *    user name.
     */
    if((getpwnam(user)) == NULL) {       /* no such login */
      (void) esprintf(error, "No such user `%s'", user);
      (void) errorMsg(error, FALSE);
    }
  }
#endif /* VALIDATE_USER_GROUP */
}

/*
 * 1. If group is a gid, verify it is a valid defined gid and then covert 
 *    it to a group name.
 * 2. If group is a group name, then verify it is a valid defined 
 *    group name.
 */
PUBLIC void validateGroup(char *group)
{
  struct group *gr;
  gid_t gid;

  /*
   * 1. If group is a gid, verify it is a valid defined gid and then covert 
   *    it to a group name.
   */
  if (isanumber(group)) {
    gid = (gid_t)atogid(group);
    gr = getgrgid(gid);
    if (gr == NULL) {                 /* no such gid */
#if (VALIDATE_USER_GROUP == TRUE)      
      (void) esprintf(error, "No such group id `%lu'", (long unsigned)gid);
      (void) errorMsg(error, FALSE);
#endif /* VALIDATE_USER_GROUP */
    }
    else {
      /* Note: assumes group has been allocated enough space already */
      (void) bzero(group, MAXGROUPNAMESIZE);
      (void) strncpy(group, gr->gr_name, MAXGROUPNAMELEN);
    }
  }
#if (VALIDATE_USER_GROUP == TRUE)      
  else {
    /*
     * 2. If group is a group name, then verify it is a valid defined 
     *    group name.
     */
    gr = getgrnam(group);
    if (gr == NULL) {                /* no such group name */
      (void) esprintf(error, "No such group `%s'", group);
      (void) errorMsg(error, FALSE);
    }
  }
#endif /* VALIDATE_USER_GROUP */
} /* validateGroup */

/*
 * Determine the current working directory. getcwd() can fail if the user 
 * does not have read or search permissions in the directory where runas is 
 * invoked from, so change to root  (who has permissions, regardless 
 * of the directory permissions) before the getcwd() call.  Always returns
 * the current working directory, otherwise the program exits.
 *
 * This function will malloc necessary storage and return a pointer to 
 * dirname.
 */
PUBLIC char *getCurrentWorkingDirectory(uid_t myuid, uid_t myeuid)
{
  char *cwd;                    /* current working directory        */
  size_t path_size;             /* size of absolute pathname buffer */
  size_t cwd_len;               /* length of NULL terminated path   */

#if (HAVE_SETREUID == 1) || (HAVE_SETRESUID == 1)
  if (setuid(0) == -1) { /* allow this process to read any directory */
    (void) esprintf(error, "%s: \"%s\", line %i setuid failed", label, __FILE__, __LINE__);
    (void) errorPMsg(error);
  }
#endif /* HAVE_SETREUID || HAVE_SETRESUID */

  cwd = __pathalloc(&path_size);
  cwd_len = path_size - 1;       /* need 1 byte for NULL                  */
  errno = 0;
  (void) getcwd(cwd, cwd_len);   /* Get the current working directory     */

  /* Error handling */
  if ((errno == ENAMETOOLONG) ||   /* Directory longer than PATH_MAX bytes  */
      (errno == ERANGE)) {
    (void) free(cwd);
    (void) errorPMsg("current working directory path too long");
  }
  if (errno == EACCES) {
    (void) free(cwd);
    (void) errorPMsg("current working directory path is not accessable");
  }

#if (HAVE_SETREUID == 1)
  if (setreuid(myuid, myeuid) == -1) {  /* change process back to user, suid root */
    (void) esprintf(error, "%s: \"%s\", line %i setreuid failed", label, __FILE__, __LINE__);
    (void) errorPMsg(error);
  }
#else
#  if (HAVE_SETRESUID == 1)
  if (setresuid(myuid, myeuid, -1) == -1) {  /* change process back to user, suid root */
    (void) esprintf(error, "%s: \"%s\", line %i setresuid failed", label, __FILE__, __LINE__);
    (void) errorPMsg(error);
  }
#  endif /* HAVE_SETRESUID */ 
#endif /* HAVE_SETREUID */
  return(cwd);
} /* getCurrentWorkingDirectory */

/*
 * 1. Return the basename of filename. 
 */
PUBLIC char *basename(const char *filename)
{
  char *base; /* basename of filename */

  /* 
   * 1. Return the basename of filename. 
   */
  base = strrchr(filename, (int)'/');
  if (base == NULL) {
    base = filename;
  } else {
    base += 1;
  }
  return (base);
}


#ifdef IGNORE_PREFIX_IN_PATH
/* 
 * 1. Search for path within the list of paths to ignore.
 *
 * Returns TRUE if path should be ignored, FALSE if it's OK. 
 */
PRIVATE BOOLEAN ignorePATH(char *path) {
  char *prefix, *prefixes, *prefixes_head;

  prefixes_head = prefixes = strdup(IGNORE_PREFIX_IN_PATH);
  prefix = strtoken(&prefixes, ':');

  /*
   * 1. Search for path within the list of paths to ignore.
   */
  while (prefix != NULL) {
    if (strncmp(prefix, path, strlen(prefix)) == 0) {
      free(prefixes_head);
      return TRUE;
    }
    prefix = strtoken(&prefixes, ':');
  }
  free(prefixes_head);
  return FALSE;
}
#endif /* IGNORE_PREFIX_IN_PATH */

/*  
 * A version of strtok() that doesn't clobber pointers when multiple
 * calls are interlaced.  
 *
 * Given a string and a delimiter, returns a NULL delimited string
 * to the token.  
 *
 * Note: This function will modify string.
 */
PUBLIC char *strtoken(char **string, char delimiter) {
  char *token = *string; 

  while(*string) {
    if (**string == '\0') {
      if (strlen(token))
	return token;
      else
	return NULL;
    }
    if (**string == delimiter) {
      **string = '\0';
      (*string)++;
      return token;
    }
    (*string)++;
  }
  return NULL;
}

/*
 * Traverses a list of usernames to determine if filename (owned by "owner")
 * should be run.
 *
 * Returns TRUE if filename should be run, FALSE otherwise.
 */
PRIVATE BOOLEAN *execthisfile(uid_t owner) {
  struct passwd *pw;
  char **okowners = ownedby;

  if (okowners != NULL) { 
    while (*okowners != NULL) {
      pw = getpwnam(*okowners);
      if (pw == NULL) {
	/* user does not exist */
	(void) esprintf(error, "%s: \"%s\", line %i %s", label, __FILE__, __LINE__, *okowners);
	(void) errorPMsg(error);
      }
      if (pw->pw_uid == owner)
	return TRUE;
      okowners++;
    }
  }
  return FALSE;
}

/*
 * Simulates the which(1) program to determine the absolute pathname of
 * a filename.  
 *
 * Returns the absolute filepath for the filename given.  Returns the
 * same filepath if the filename is a relative filepath. 
 * Calls errorMsg if no match is found.
 *
 * char *filename;                  name of file to find absoulte path of
 */
PUBLIC char *which(const char *filename)
{
  size_t filename_len;           /* size of NULL terminated filename string */
  char *envptr = NULL;           /* pointer to environmental variable value */
  char *tokenvptr;               /* copy of envptr                          */
  char *tokenvptr_head;          /* initial tokenvptr value                 */
  char *token;                   /* strtok() pointer to tokenvptr           */
  size_t token_len;              /* size of NULL terminated PATH element    */
  char *path;                    /* absolute path of filename               */
  size_t path_size;              /* size of absolute pathname buffer        */
  size_t path_len;               /* length of NULL terminated path          */
  struct stat st;                /* path file information                   */
  BOOLEAN path_found = FALSE;    /* set to TRUE if filepath is valid        */

  /* 
   * 1. Return filename if filename is NULL, an absolute or relative 
   *    filepath.
   *    IGNORE_SYMLINKS: In addition, return filename if it is not
   *    a symbolic link.
   */
  if ((filename == NULL) ||               /* no pathname       */
      (strncmp("/", filename, 1) == 0) || /* absolute pathname */
      (strncmp(".", filename, 1) == 0)) { /* relative pathname */
#if (IGNORE_SYMLINKS == TRUE)
    if (lstat(filename, &st) != 0) {
      (void) esprintf(error, "%s: \"%s\", line %i %s", label, __FILE__, __LINE__, filename);
      (void) errorPMsg(error);
    }
    if (((st.st_mode) & S_IFMT) == S_IFLNK) { /* file a link? */
      (void) fprintf(stderr, "%s: %s is a symlink and will not be executed.\n", label, filename);
      (void) exitOK();
    }
#endif /* IGNORE_SYMLINKS */
    if (stat(filename, &st) != 0) {
      if (errno == ENOENT)
	(void) esprintf(error, "%s: %s", label, filename);
      else
	(void) esprintf(error, "%s: \"%s\", line %i %s", label, __FILE__, __LINE__, filename);
      (void) errorPMsg(error);
    }
    if ((st.st_mode & S_IWOTH) == S_IWOTH) {
      /* File is world writeable.  Do not run it. */
      (void) fprintf(stderr, "%s: %s is world writeable and will not be executed.\n", label, filename);
      (void) exitOK();
    }
#if (EXEC_ONLY_ADMIN_OWNED_FILES == TRUE)
    if (! execthisfile(st.st_uid)) {
      (void) fprintf(stderr, "%s: %s is not owned by an administrative account\n  and will not be executed.\n", label, filename);
      (void) exitOK();
    }
#endif /* EXEC_ONLY_ADMIN_OWNED_FILES */
    if (which_command)
      printf("File to run: %s\n", filename);
    return(filename);
  }

  filename_len = strlen(filename);

  envptr = getenv("PATH");
  if (envptr != NULL) { /* No $PATH? PATH exists even through a pipe! */
    tokenvptr = tokenvptr_head = strdup(envptr); /* muck with copy of **environ pointer */
    token = strtok(tokenvptr, ":");
    path = __pathalloc(&path_size);
    path_len = path_size - 1;
    while ((token != NULL) && (!path_found)){
#ifdef IGNORE_PREFIX_IN_PATH
      if (ignorePATH(token) == FALSE) {
    /* Ignore "." (current working directory) if it is in PATH */
#endif /* IGNORE_PREFIX_IN_PATH */
	
	if (stat(token, &st) == 0) { /* does the directory exist? */
	  if ((st.st_mode & S_IWOTH) == S_IWOTH) { 
	    /* directory is world writeable */
	    if (which_command)
	      printf("Skipping world writeable $PATH directory: %s\n", token);
	  } else if ((st.st_mode & S_IWGRP) == S_IWGRP) { 
	    /* directory is group writeable */
	    if (which_command)
	      printf("Skipping group writeable $PATH directory: %s\n", token);
	  }
#if (EXEC_ONLY_ADMIN_OWNED_FILES == TRUE)
	  else if (! execthisfile(st.st_uid)) {
	    if (which_command)
	      printf("Skipping non-administrator owned directory: %s\n", token);
	  }
#endif
	  else { /* $PATH is ok, now check the file */

	    token_len = strlen(token);
	    if ((token_len + filename_len + 1) <= path_len) {
	      (void) bzero(path, path_size);
	      (void) strncpy(path, token, token_len);
	      (void) strncat(path, "/", 1);
	      (void) strncat(path, filename, filename_len);
	      if (stat(path, &st) == 0) { /* does the file exist? */ 
		path_found = TRUE;
#if (EXEC_ONLY_ADMIN_OWNED_FILES == TRUE)
		if (! execthisfile(st.st_uid)) {
		  path_found = FALSE;
		  if (which_command) 
		    printf("Skipping non-administrator owned file: %s\n", path);
		} else {
#endif /* EXEC_ONLY_ADMIN_OWNED_FILES */
		  if ((st.st_mode & S_IWOTH) == S_IWOTH) {
		    /* File is world writeable.  Do not run it. */
		    path_found = FALSE;
		    if (which_command)
		      (void) printf("Skipping world writeable file: %s\n", path);
		  }
#if (IGNORE_SYMLINKS == TRUE)
		  if (lstat(path, &st) != 0) {
		    (void) esprintf(error, "%s: \"%s\", line %i %s", label, __FILE__, __LINE__, path);
		    (void) errorPMsg(error);
		  }
		  if (((st.st_mode) & S_IFMT) == S_IFLNK) { /* file a link? */
		    path_found = FALSE;
		    if (which_command) 
		      printf("Skipping link: %s\n", path);
		  }
#endif /* IGNORE_SYMLINKS */
#if (EXEC_ONLY_ADMIN_OWNED_FILES == TRUE)
		}
#endif /* EXEC_ONLY_ADMIN_OWNED_FILES */
	      } /* does the file exist? */
	    } /* ignores paths longer than the size of the absolutepath buffer */
	  }
	} /* does $PATH directory exist? */
      
#ifdef IGNORE_PREFIX_IN_PATH
      } else if (which_command) 
	printf("Skipping executable files in %s\n", token);
#endif /* IGNORE_PREFIX_IN_PATH */
      token = strtok(NULL, ":"); 
    }

    (void) free(tokenvptr_head);
    if (path_found) {
      if (which_command)
	printf("File to run: %s\n", path);
      return(path);
    }
  }
  (void) esprintf(error, "%s: file not found in $PATH", filename);
  (void) errorMsg(error, FALSE);
  return(NULL);  /* Satisfies control reaching end of non-void function  */
} /* which */

#if (USE_ACF == TRUE)
/*
 * 1. Return TRUE if case insensitive string = yes.
 * 2. Return FALSE otherwise.
 */
PUBLIC int convertToBoolean(const char *string)
{
  /*
   * 1. Return TRUE if case insensitive string = yes.
   */
  if (strncasecmp(string, "yes", (size_t)3) == 0) {
    return(TRUE);
  } else {
    /*
     * 2. Return FALSE otherwise.
     */
    return(FALSE);
  }
} /* convertToBoolean */

/*
 * 1. Display an error message that the command line arguments 
 *    are being removed.
 * 2. Remove arguments 1 through (argc-1).
 */
PUBLIC void removeOptions(int *argc, char **argv)
{
  int x;            /* loop counter                               */
  int c;            /* number of command line arguments to remove */

  c = (*argc - 1); /* remove all arguments past argv[0]           */

  /*
   * 1. Display an error message that the command line arguments 
   *    are being removed.
   */
  if (c > 0) {
    (void) fprintf(stderr, "%s: Command line arguments will not be passed.\n", label);

    /* 
     * 2. Remove arguments 1 through (argc-1).
     */
    for (x=1; x <= c; x++){ /* remove arguments 1 through (argc-1)         */
      argv[x] = NULL;       /* make it NULL                                */
      (*argc)--;            /* total number of command line arguments left */
    }
  }
} /* removeOptions */

/*
 * 1. Remove all leading & trailing whitespace from a non-NULL string.
 */
PUBLIC void cleanString(char *string)
{
  char *buff;                           /* temporary storage string          */
  size_t string_len;                    /* non-NULL terminated string length */
  size_t string_size;                   /* NULL terminated string size       */

/*
 * 1. Remove all leading & trailing whitespace from a non-NULL string.
 */
  if (string != NULL) {                    /* ignore NULL strings            */
    string_len = strlen(string);           /* length of the string to copy   */
    string_size = string_len + 1;          /* +1 for NULL                    */
    buff = __stralloc(string_size);        /* allocate space for temp string */
    (void) sscanf(string, "%s", buff);     /* remove whitespace from string  */
    (void) bzero(string, string_len);      /* empty buffer before overwriting*/
    (void) strncpy(string, buff, string_len); /* copy buffer back to string  */
    (void) free(buff);                     /* release memory back            */
  }
} /* cleanString */
#endif /* USE_ACF */

/*
 * 1. Allocate memory, keeping the origin of __FILE__ and __LINE__.
 * 2. Fill the string with NULLs.
 * 3. Return the address of the allocated memory.
 *
 * size_t size;        The amount of memory to allocate
 * const char *file;   File name where stralloc call occurred
 * int line;           Line number of file where stralloc call occurred 
 */
PUBLIC char *stralloc(size_t size, const char *file, int line)
{
  char *addr;      /* allocated memory                                 */
 
  /* 
   * 2. Allocate memory, keeping the origin of __FILE__ and __LINE__.
   */
  addr = (char *)errorCheckMalloc(size, file, line);

  /*
   * 2. Fill the string with NULLs.
   */
  (void) bzero(addr, size);

  /*
   * 3. Return the address of the allocated memory.
   */
  return(addr);
} /* stralloc */

/*
 * 1. If maxpathlen is 0, then set maxpathlen to a large arbitrary value.  
 *    HAVE_PATHCONF: Attempt to determine the run-time POSIX value of 
 *    the maximum absolute path length supported.
 * 2. Allocate memory, keeping the origin of __FILE__ and __LINE__.
 * 3. Return the address of the allocated memory.
 *
 * size_t *size;       Size of allocated buffer
 * const char *file;   File name where call pathalloc occurred
 * int line;           Line number of file where pathalloc call occurred 
 */
PUBLIC char *pathalloc(size_t *size, const char *file, int line)
{
  char *addr;      /* allocated memory                                  */

  /*
   * 1. If maxpathlen is 0, then set maxpathlen to a large arbitrary value.  
   *    HAVE_PATHCONF: Attempt to determine the run-time POSIX value of 
   *    the maximum absolute path length supported.
   *    This logic is skipped if the global static maxpathlen variable
   *    is set to a non-zero value.
   */
  if (maxpathlen == 0) { /* first time through */
    if ((maxpathlen = pathconf("/", _PC_PATH_MAX)) == -1) {
      if (errno == 0) { /* indeterminate value */
	maxpathlen = RUNAS_PATH_MAX;
      } else {          /* error occured */
	(void) errorPMsg("pathconf(_PC_PATH_MAX) failed");
      }
    } else {
      maxpathlen++; /* account for "/" directory length */
    } /* if pathconf() == -1 */
  } /* if maxpathlen == 0 */

  /* 
   * 2. Allocate memory, keeping the origin of __FILE__ and __LINE__.
   */
  *size = maxpathlen + 1;  /* add 1 for NULL */
  addr = (char *) stralloc(*size, file, line); 

  /*
   * 3. Return the address of the allocated memory.
   */
  return(addr);
} /* pathalloc */

/*
 * 1. Allocate memory.
 * 2. Display error and exit if memory cannot be allocated. 
 * 3. Return the address of the allocated memory.
 *
 * size_t size;        The amount of memory to allocate
 * const char *file;   File name where function call occurred
 * int line;           Line number of file where function call occurred 
 */
PUBLIC void *errorCheckMalloc(size_t size, const char *file, int line)
{
  void *addr;      /* allocated memory                                 */

  /* 
   * 1. Allocate memory.
   */
  errno = 0;
  addr = malloc(size);

  /*
   * 2. Display error and exit if memory cannot be allocated. 
   */
  if (addr == NULL) { /* ENOMEN */
    (void) esprintf(error, "%s: \"%s\", line %i malloc failed, size=%li", label, file, line, (long)size);
    (void) errorPMsg(error);
  } /* if (addr == NULL) */

  /*
   * 3. Return the address of the allocated memory.
   */
  return(addr); 
} /* errorCheckMalloc */

/*
 * 1. Return TRUE if string is a number, otherwise return FALSE.
 *
 * Note: does not deal with signed ("+" or "-") strings.
 */
PUBLIC BOOLEAN isanumber(const char *s) {
  /*
   * 1. Return TRUE if string is a number, otherwise return FALSE.
   */ 
  if (s == NULL) {
    return(FALSE);
  }
  while (strlen(s)) {
    if (!(isdigit((int)*s++))) {
      return(FALSE);
    }
  }
  return(TRUE);
} /* isanumber */

/*
 * 1. Returns TRUE if shell is a restricted shell (as defined in 
 *    /etc/shells) otherwise returns FALSE.
 */
PUBLIC BOOLEAN isrshell(const char *shell)
{
  char *entry; /* line entry from getusershell(). */
  /*
   * 1. Returns TRUE if shell is a restricted shell (as defined in 
   *    /etc/shells) otherwise returns FALSE.
   */
  if (shell != NULL) {
    (void) setusershell();
    while ((entry = getusershell()) != NULL) {
      if ((*entry != '#') && (strcmp(entry, shell) == 0)) {
	(void) endusershell();
	return(TRUE);
      }
    }
    (void) endusershell();
  } /* shell != NULL */
  return(FALSE);
}

#if (HAVE_GETUSERSHELL != 1)
/*
 * 1. Open the /etc/shells file if not already open or set the pointer
 *    to the beginning of the list. 
 * 2. If /etc/shells is already open, then rewind the file pointer to the
 *    beginning of the file.
 *
 * Note: this function is not reentrant.
 */
PUBLIC void setusershell(void)
{
  if (fpShells == NULL) {
    /* 
     * 1. Open the /etc/shells file if not already open or set the pointer
     *    to the beginning of the list. 
     */
    fpShells = fopen(ETCSHELLS, "r");
    if ((fpShells == NULL) && (etcshells == NULL)) 
      etcshells = shells;
  } else 
    /* 
     * 2. If /etc/shells is already open, then rewind the file pointer to the
     *    beginning of the file.
     */
    (void) rewind(fpShells);
}

/*
 * 1. Close the /etc/shells file if it exists and is open.
 *
 * Note: this function is not reentrant.
 */
PUBLIC void endusershell(void)
{
  /*
   * 1. Close the /etc/shells file if it exists and is open.
   */
  if (fpShells != NULL)
    fclose(fpShells);
}

/*
 * 1. Sequentially returns one line from the /etc/shells file.
 *    Returns NULL on EOF.
 * 2. Sequentially returns on entry from a list of default shell filepaths
 *    if /etc/shells does not exist.
 *
 * Note: this function is not reentrant.
 */
PUBLIC char *getusershell(void)
{
  static char line[MAXIMUMPATHLEN];
  /*
   * 1. Sequentially returns one line from the /etc/shells file.
   *    Returns NULL on EOF.
   */
  if (fpShells) {
    if(fgets(line, sizeof(line)-1, fpShells) == NULL)
      return NULL;
    /* Need to remove \n from line before returning. */
    if (line[strlen(line)-1] == '\n')
      line[strlen(line)-1] = '\0';
    return line;
  } else {
    /*
     * 2. Sequentially returns on entry from a list of default shell filepaths.
     *    if /etc/shells does not exist.
     */
    if (shells == etcshells) {
      etcshells++;
      return *shells;
    } else if (*etcshells == NULL) {
      return NULL;
    }
    return *(etcshells++);
  }
}
#endif /* HAVE_GETUSERSHELL */
/* end misc.c */
