/* syslog.c */

/* 
 * Determine the maximum length that syslog() can handle before it 
 * core dumps.  Since syslog() does not return a value or generate
 * an errno if it fails, the only way to determine the maximum length
 * is to manually correlate the last syslog message to appear in a logfile
 * with the message information printed for each call to syslog.
 * Must manually interrupt to stop.
 *
 * Add the following to /etc/syslog.conf in order to monitor the
 * syslog messages:
 
local0.debug       /var/log/test
 
 */

/* System headers */
#include <stdio.h>
#include <stdlib.h>
#include <sys/utsname.h>
#include <syslog.h>
#include <string.h>
#include <unistd.h>

/* Macros */
#define SYSLOG_FACILITY LOG_LOCAL0
#define SYSLOG_PRIORITY LOG_DEBUG
#define SYSLOG_OPTIONS  LOG_NDELAY

#define STARTLEN    128  /* initial syslog buffer size */
#define BOUNDARY    128  /* limitations seem to occur on 128 byte boundaries */
#define DATETIMELEN 16   /* length of date and time, including spacing       */

/* Some older platforms do not have these defined */
#ifndef TRUE
#define TRUE         1
#endif
#ifndef FALSE
#define FALSE        0
#endif

#define DEBUG

/* Main */
int main(argc, argv, envp) 
int argc;
char *argv[];
char *envp[];
{

  char *format;               /* syslog message                    */
  int formatlen = 0;          /* syslog format length              */
  struct utsname *host;       /* uname information                 */
  int hostnamelen;            /* strlen of hostname                */
  char *label;                /* syslog label                      */
  int  labelen;               /* strlen of label                   */ 
  int totalen;                /* total strlen of syslog message    */
  int chr = 65;               /* "A"; character to populate memset */

  host = (struct utsname*)malloc(sizeof(struct utsname));
  if (host == NULL) { /* ENOMEM */
    (void) perror("malloc");
  }
  /* Get host information */
  if (uname(host) == -1) {
    (void) perror("uname");
  } 
  hostnamelen = strlen(host->nodename);
  (void) free(host);

  /* use the name of the executable as the label name */
  label = argv[0];
  labelen = strlen(label);

  /* 3 bytes account for a colon and two spaces */
  formatlen = (STARTLEN - DATETIMELEN - labelen - hostnamelen - 3); 
  (void) openlog(label, SYSLOG_OPTIONS, SYSLOG_FACILITY);
  while (TRUE) {
    format = (char *)malloc(formatlen);
    if (format == NULL) { /* ENOMEM */
      (void) perror("malloc");
    }
    (void) memset(format, chr, formatlen);  /* populate the address space */
    (void) syslog(SYSLOG_PRIORITY, "%s", format);
    /* 16 chars for date and time, including spacing: "Jan 23 20:30:41 " */
    /* hostname + space + label + colon + space + message */
    totalen = DATETIMELEN + hostnamelen + 1 + labelen + 2 + formatlen;
#ifndef DEBUG
    (void) printf("Msg Fill: \"%c\". #define SYSLOG_BUFSIZE %d\n", chr++, totalen);
#else
    (void) printf("Msg Fill: \"%c\". Msg length: %4d. Label length: %d.  #define SYSLOG_BUFSIZE %d\n", chr++, formatlen, labelen, totalen); 
#endif /* DEBUG */
    (void) free(format);
    (void) sleep(1);
    formatlen += BOUNDARY;
  }
  (void) closelog();
}

/* end syslog.c */
