/* runas.c -- main functions
 *
 * $Id: runas.c,v 1.38 1998/11/10 23:26:53 ccarp Exp $
 * 
 * runas - execute a single command as a different user and/or group
 * Copyright (C) 1998 Chris Carpinello
 *
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 * Send all bugs, changes, and problems to the maintainer listed in
 * "runas --info" output.
 */

/* System headers */
#include <stdio.h>
#if (HAVE_STDLIB_H == 1)
#include <stdlib.h>
#endif /* HAVE_STDLIB_H */
#include <pwd.h>
#include <grp.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <ctype.h>
#if (HAVE_LIMITS_H == 1)
#include <limits.h>
#endif /* HAVE_LIMITS_H */
#if (HAVE_MALLOC_H == 1)
#include <malloc.h>
#endif /* HAVE_MALLOC_H */
#include <string.h>
#include <errno.h> 
#if (HAVE_UNISTD_H == 1)
#include <unistd.h>
#endif /* HAVE_UNISTD_H */
#include <fcntl.h>

/* Local headers and macros */
#include "defines.h"
#include "config.h"
#include "acf.h"
#include "log.h"

/* Global scope variables */
PUBLIC char error[ERRORSIZE];  /* error message buffer for printf formatting */

PUBLIC BOOLEAN which_command = FALSE;   /* --which specified? */
PUBLIC BOOLEAN help_command = FALSE;    /* --help specified? */
PUBLIC BOOLEAN version_command = FALSE; /* --version specified? */
PUBLIC BOOLEAN info_command = FALSE;    /* --info specified? */
PUBLIC BOOLEAN access_command = FALSE;  /* --access specified? */

#ifdef MAXIMUMPATHLEN
PUBLIC size_t maxpathlen = MAXIMUMPATHLEN;
#else
PUBLIC size_t maxpathlen = 0;
#endif /* MAXIMUMPATHLEN */

PUBLIC char *label; /* the name runas is invoked as */

/*
 * Ok_Folks contains a list of uids that bypass syslog when runas is 
 * invoked (only root should be in here - if any at all),
 */
#if (USE_SYSLOG == TRUE)
PUBLIC uid_t Ok_Folks[] = { 
#ifdef OK_FOLKS
OK_FOLKS, 
#endif
OK_FOLKS_DELIMITER };
#endif /* USE_SYSLOG */

/* 
 * File scope variables
 */

/* This is set to TRUE if "command" is a restricted shell */ 
PRIVATE BOOLEAN restricted_shell = FALSE; 

/* Determines if the environment is preserved or sanitized for each command */
PRIVATE BOOLEAN preserve_env = PRESERVE_ENVIRONMENT_DEFAULT;

PRIVATE char *sysname = SYS;

#if (USE_SYSLOG == TRUE)
PRIVATE char runas_syslog[] = "Syslog support enabled";
#else
PRIVATE char runas_syslog[] = "Syslog support disabled";
#endif /* USE_SYSLOG */

#if (USE_SECURE_PATH == TRUE)
PRIVATE char runas_securepath[] = "Secure PATH enabled";
#else
PRIVATE char runas_securepath[] = "Secure PATH disabled.";
#endif /* USE_SECURE_PATH */

#if (VALIDATE_USER_GROUP == TRUE)
PRIVATE char runas_validate_user_group[] = "Process must be run as a user/group with a valid passwd/group file entry.";
#else
PRIVATE char runas_validate_user_group[] = "Process may be run as any uid and gid.";
#endif /* VALIDATE_USER_GROUP */

#ifdef IGNORE_PREFIX_IN_PATH
PRIVATE char runas_ignorepath[] = "Directories in $PATH with the following prefixes will be ignored:";
#endif /* IGNORE_PREFIX_IN_PATH */

#if (IGNORE_SYMLINKS == TRUE)
PRIVATE char runas_symlinks[] = "Symlinks in $PATH are ignored.";
#endif /* IGNORE_DOTPATH */

#if (EXEC_ONLY_ADMIN_OWNED_FILES == TRUE)
PRIVATE char runas_exec_admin_files[] = "Only files owned by the following usernames will be executed:";
#endif /* EXEC_ONLY_ADMIN_OWNED_FILES */

#if (ALLOW_ROOT_SHELLS == FALSE) && (EXEC_ONLY_ADMIN_OWNED_FILES == TRUE)
PRIVATE char runas_allow_root_shells[] = "The following shells cannot be run as root:";
#else
PRIVATE char runas_allow_root_shells[] = "Root shells are allowed.";
#endif /* ALLOW_ROOT_SHELLS */

#if (USE_ACF == TRUE)
#if (HAVE_SETNETGRENT == 1)
PRIVATE char runas_netgroups[] = "Access control file netgroup support enabled.";
#else
PRIVATE char runas_netgroups[] = "Access control file netgroup support disabled.";
#endif /* HAVE_SETNETGRENT */
#if (TIGHT_CHECKING == TRUE)
PRIVATE char runas_tight[] = "Tight checking enforcement of access control file enabled";
PRIVATE char runas_tight2[] = "Tight checking enforcement of commands within access control file enabled";
#else
PRIVATE char runas_tight[] = "Tight checking enforcement of access control file disabled.";
PRIVATE char runas_tight2[] = "Tight checking enforcement of commands within access control file disabled.";
#endif /* TIGHT_CHECKING */
#endif /* USE_ACF */

PRIVATE char runas_preserve_env[] = "The current shell environment is preserved when running a command.";
PRIVATE char runas_sanitize_env[] = "The current shell environment is replaced by a minimum shell environment\n      when running a command.";

/* External variables */
#include "extern.h"

/* External function prototypes */ 
#include "proto.h"

/* Local function prototypes */
PRIVATE void checkAccess       __P((uid_t myuid, uid_t myeuid, gid_t mygid, int *argc, char **argv, char *runuidptr, char *rungidptr, BOOLEAN showCommands, BOOLEAN *needACF));
PRIVATE void infoAccess        __P((uid_t myuid, gid_t mygid, BOOLEAN needACF));
PRIVATE void infoBugReport     __P((uid_t myuid, gid_t mygid, BOOLEAN needACF));
PRIVATE void infoCompile       __P((void));
PRIVATE void infoUsage         __P((void));
PRIVATE void infoVersion       __P((void));
PRIVATE void parseUserGroupArg __P((char *arg, uid_t myuid, char *runuidptr, char *rungidptr));
PRIVATE void setSecurePATH     __P((void));
PRIVATE void setEnvironment    __P((struct passwd *passwd, const char *shell, BOOLEAN needACF)); 
PRIVATE void setIdentity       __P((char *runuidptr, const char *rungidptr, int argc, char **argv, uid_t myuid, uid_t myeuid, BOOLEAN setID, BOOLEAN needACF));

/* Functions */

/*
 * setIdentity
 *
 *  1. Set the default user to run this process as to root if a user has
 *     not already been specified. 
 *  2. Convert specified user name to a uid.
 *  3. If the group is not defined, run the process as the base group of 
 *     the user specified, otherwise convert the specified groupname to a 
 *     gid. If the uid the process will be run as is 0, then for 
 *     administrative purposes, set the gid of the process to ROOT_GID 
 *     instead. If "-root" is explicitly given as the user to run this 
 *     process as, then root's base group id will be used instead of 
 *     ROOT_GID. 
 *  4. Set a secure PATH. 
 *  5. Determine the absolute path of the command to run and replace
 *     the unqualified command path with the absolute command path.
 *  6. Determine if the command specified is a restricted shell.
 *  7. USE_SYSLOG: Log the command and command arguments.
 *  8. Set environment variables (except for PATH). 
 *  9. Give access to supplemental groups.
 * 10. Set the gid and uid of this process.
 *
 * char  *runuidptr;        user name                             
 * const char  *rungidptr;  group name                            
 * int   argc;              number of command line arguments      
 * char  **argv;            command line arguments                
 * uid_t myuid;             uid of the user invoking runas        
 * uid_t myeuid;            uid of the user invoking runas        
 * BOOLEAN setID;           was user.group specified in argv?     
 * char *environ[];         process environmental variables       
 * BOOLEAN needACF;         access control file needed for access?
 */

PRIVATE void setIdentity(char *runuidptr, 
			 const char *rungidptr, 
			 int argc, 
			 char **argv, 
			 uid_t myuid, 
			 uid_t myeuid, 
			 BOOLEAN setID, 
			 BOOLEAN needACF)
{
  struct passwd	*pw;             /* passwd entry for myuid               */ 
  struct passwd passwd;          /* copy of passwd struct                */
  struct group *group = NULL;    /* group entry of rungidptr             */
  uid_t runuid;                  /* uid of runuidptr                     */
  gid_t rungid;                  /* gid of rungidptr                     */

  /*
   *  1. Set the default user to run this process as to DEFAULT_UID if a 
   *     user has not already been specified. 
   */
  if (!strlen(runuidptr)) {
    (void) snprintf(runuidptr, MAXUSERNAMELEN, "%lu", (long unsigned)DEFAULT_UID);
    pw = getpwuid((uid_t)DEFAULT_UID);
    if (pw == NULL) { /* No such user or ENOMEM */
      (void) esprintf(error, "%s: \"%s\", line %i getpwuid failed", label, __FILE__, __LINE__);
      (void) errorPMsg(error);
    } 
    /* 
     * which() makes a call to execthisfile() which in turn calls getpwnam()
     * which is overwriting the static area used by getpwnam.  We make our 
     * own copy to avoid the clobbering. 
     */
    passwd.pw_name = strdup(pw->pw_name);
    passwd.pw_dir = strdup(pw->pw_dir);
    passwd.pw_shell = strdup(pw->pw_shell);
    passwd.pw_uid = pw->pw_uid;
    passwd.pw_gid = pw->pw_gid;
    runuid = pw->pw_uid;
  } else { 

    /* 
     *  2. Convert specified user name to a uid. 
     */
    pw = getpwnam(runuidptr);
    if (pw == NULL) { /* No such user or ENOMEM */
      if (errno != ENOMEM) { /* No such user */
#if (VALIDATE_USER_GROUP == TRUE)
	(void) fprintf(stderr, "%s: user (%s) does not have a passwd file entry.\n", label, runuidptr);
	(void) exitOK();
#else
	if (!(isanumber(runuidptr))) { /* Not a uid specified */
	  (void) fprintf(stderr, "%s: No such user ``%s''.\n", label, runuidptr);
	  (void) exitOK();
	} 
#endif /* VALIDATE_USER_GROUP */
      } else { /* ENOMEM */
	(void) esprintf(error, "%s: \"%s\", line %i getpwnam failed", label, __FILE__, __LINE__);
	(void) errorPMsg(error);
      }
    } else { /*valid user */
      /* 
       * which() makes a call to execthisfile() which in turn calls getpwnam()
       * which is overwriting the static area used by getpwnam.  We make our 
       * own copy to avoid the clobbering. 
       */
      passwd.pw_name = strdup(pw->pw_name);
      passwd.pw_dir = strdup(pw->pw_dir);
      passwd.pw_shell = strdup(pw->pw_shell);
      runuid = passwd.pw_uid = pw->pw_uid;
      passwd.pw_gid = pw->pw_gid;
    }
  }
  
  /*
   *  3. If the group is not defined, run the process as the base group of 
   *     the user specified, otherwise convert the specified groupname to a 
   *     gid. If the uid the process will be run as is 0, then for 
   *     administrative purposes, set the gid of the process to ROOT_GID 
   *     instead. If "-root" is explicitly given as the user to run this 
   *     process as, then root's base group id will be used instead of 
   *     ROOT_GID. 
   *
   *     Note: It is certainly possible for an insecure system to have a user
   *     with a uid of 0 that is not user "root".  Runas makes no attempt to
   *     distinquish or equate usernames "root", "operator", "t00r" or
   *     other usernames to each other -- all runas cares about is that the 
   *     uid is 0.
   */
  if (!strlen(rungidptr)) {
    rungid = passwd.pw_gid;
#ifdef ROOT_GID
    if ((runuid == 0) && (setID == FALSE)) {
      rungid = ROOT_GID; 
    }
#endif /* ROOT_GID */
  } else {
    /* Convert the specified group name to a gid */
    
    group = getgrnam(rungidptr);
    if (group == NULL) {
      if (errno == 0) { /* No such group */
#if (VALIDATE_USER_GROUP == TRUE)
	(void) fprintf(stderr, "%s: group (%s) does not have a group file entry.\n", label, rungidptr);
	(void) exitOK();
#else
	if (!(isanumber(rungidptr))) { /* Not a gid specified */
	  (void) fprintf(stderr, "%s: No such group ``%s''.\n", label, rungidptr);
	  (void) exitOK();
	} 
	rungid = (gid_t) atogid(rungidptr);
#endif /* VALIDATE_USER_GROUP */
      } else { /* ENOMEM */
	(void) esprintf(error, "%s: \"%s\", line %i getgrnam failed", label, __FILE__, __LINE__);
	(void) errorPMsg(error);
      }
    } else { /* valid group */
      rungid = group->gr_gid;
    }
  }

  /* 
   *  4. Set a secure PATH. 
   */
  setSecurePATH();

  /* 
   *  5. Determine the absolute path of the command to run and replace
   *     the unqualified command path with the absolute command path.
   */
  argv[0] = which(argv[0]); 

  /* 
   *  6. Determine if the command specified is a restricted shell.
   */
  restricted_shell = isrshell(argv[0]);

#if (ALLOW_ROOT_SHELLS == FALSE)
  if (restricted_shell && (runuid == 0)) {
    /* No root shells allowed */
    (void) esprintf(error, "%s: Restricted shells cannot be run as root!", argv[0]);
    (void) errorMsg(error, FALSE);
  }
#endif /* ALLOW_ROOT_SHELLS */

#if (USE_SYSLOG == TRUE)
  /* 
   *  7. USE_SYSLOG: Log the command and command arguments.
   *
   * Note: This must occur before setEnvironment() as this logs the 
   * current working directory from where runas was invoked and 
   * setEnvironment() may call chdir().
   *
   * Note: Don't syslog --which usage.
   */ 
  if (! which_command)
    (void) syslogCommand(myuid, myeuid, runuid, rungid, argc, argv);
#endif /* USE_SYSLOG */

  /* 
   *  8. Set environment variables (except for PATH). 
   *
   * Note: Must occur before which(), as the PATH is modified if
   * USE_SECURE_PATH is TRUE.
   */
  (void) setEnvironment(&passwd, argv[0], needACF); 

  /* 
   *  9. Give access to supplemental groups.
   *
   * Note: This must occur before setuid() for initgroups to set the
   * groups.  
   */
#if (HAVE_INITGROUPS == 1)
  if (initgroups(&passwd.pw_name, passwd.pw_gid) == -1) {
    (void) esprintf(error, "%s: \"%s\", line %i initgroups failed", label, __FILE__, __LINE__);
    (void) errorPMsg(error);
  }
#endif

  /*
   * 10. Set the gid and uid of this process.
   * 
   * Note: setgid() must occur before setuid() when the runuid != 0, else
   * the non-root user will not have the permission needed for setgid().
   */
  if (setgid(rungid) == -1) {
    (void) esprintf(error, "%s: \"%s\", line %i setgid failed", label, __FILE__, __LINE__);
    (void) errorPMsg(error);
  }
  if(setuid(runuid) == -1) {
    (void) esprintf(error, "%s: \"%s\", line %i setuid failed", label, __FILE__, __LINE__);
    (void) errorPMsg(error);
  }

  /* free space allocated with strdup */
  free(passwd.pw_name);
  free(passwd.pw_dir);
  free(passwd.pw_shell);
} /* setIdentity */

/* 
 * 1. PATH is replaced with a secure $PATH if $PATH does not already exist
 *    or the environment is to be sanitized.  Otherwise, a secure PATH is 
 *    prepended to the existing path.
 * 2. Display the modified $PATH.
 */
void setSecurePATH(void) {
#if (USE_SECURE_PATH == TRUE)  
  char *path;           /* PATH environmental variable value */  
  char *securepath;     /* The new, secure PATH */
  int securepathsize;   /* Size of the securepath buffer */

  path = getenv("PATH");           /* retrieve existing "insecure" path */
  if ((preserve_env == FALSE) || (path == NULL)) {
    /* PATH is replaced with a secure $PATH */
    securepathsize = sizeof(SECURE_PATH);      /* sizeof includes a NULL */ 
    securepath = __ecmalloc(securepathsize);
    (void) strcpy(securepath, SECURE_PATH);
  } else {
    /* A secure PATH is prepended to the existing path. */
    if (path != NULL) {
      securepathsize = strlen(path) + sizeof(SECURE_PATH) + 1; /* 1 for ':' */
      securepath = __ecmalloc(securepathsize);
      (void) strcpy(securepath, SECURE_PATH);  /* secure path */
      (void) strncat(securepath, ":", 1);      /* PATH delimiter */
      (void) strcat(securepath, path);         /* append PATH to SECURE_PATH */
    }
  }
  /* Set the PATH environment variable. */ 
  securepath[securepathsize-1] = '\0'; /* because we're paranoid */
  (void) setenvar("PATH", securepath);
#endif
  /*
   * 2. Display the modified $PATH.
   */
  if (which_command) 
    printf("Modified $PATH = %s\n", getenv("PATH"));
}

/*
 * setEnvironment
 *
 * 1. If --sanitize is specificed, replace the current shell environment 
 *    with a sanitized environment.
 * 2. If command is a restricted shell and the admin doesn't need to use
 *    the ACF file, set SHELL to command otherwise, set SHELL to NOSHELL.
 * 3. Set a secure umask.
 *
 * struct passwd *passwd    password info of user to assume identify of
 * const char *command      the fullpath of the command to run
 * BOOLEAN needACF;         access control file needed for access?
 */
PRIVATE void setEnvironment(struct passwd *passwd, 
			    const char *command,
			    BOOLEAN needACF)
{
  char *term;           /* TERM environmental variable value */
  char *shlvl;          /* SHLVL environmental variable value */
  char *path;           /* PATH environmental variable value */

  /*
   * 1. If --sanitize is specificed, replace the current shell environment 
   *    with a sanitized environment.
   */
  if (preserve_env == FALSE) {
    term = getenv("TERM");               /* preserve TERM variable */
    shlvl = getenv("SHLVL");             /* preserve SHLVL variable */
    path = getenv("PATH");               /* preserve PATH variable */ 

    /* Create new sanitized shell environment. */
    environ = (char **) __ecmalloc(sizeof(char *) * 2);
    environ[0] = NULL;

    if (term != NULL) 
      (void) setenvar("TERM", term);     /* set TERM to preserved value  */
    if (shlvl != NULL) 
      (void) setenvar("SHLVL", shlvl);   /* set SHLVL to preserved value */
    if (path != NULL) 
      (void) setenvar("PATH", path);   /* set SHLVL to preserved value */

    /*
     * Setting LOGNAME and USER to the new user enables the new user's shell
     * configuration files to be referenced instead of the admin's files.
     */
    (void) setenvar("LOGNAME", passwd->pw_name); /* set LOGNAME variable */
    (void) setenvar("USER", passwd->pw_name);    /* set USER variable    */

    if (passwd->pw_dir)
      (void) setenvar("HOME", passwd->pw_dir);   /* set HOME variable    */

    /* 
     * Change working directory to the new user's home directory only if
     * the command is a restricted shell. 
     */
    if (restricted_shell) {
      if (chdir(passwd->pw_dir) != 0) {
	(void) esprintf(error, "%s: Cannot change directory to %s", label, passwd->pw_dir);
	(void) errorPMsg(error);
      }
    }
  }

  /* 
   * 2. If command is a restricted shell and the admin doesn't need to use
   *    the ACF file, set SHELL to command otherwise, set SHELL to NOSHELL.
   */ 
  if ((restricted_shell) && (needACF == FALSE))   /* set SHELL variable   */
    (void) setenvar("SHELL", command);         
  else 
    (void) setenvar("SHELL", NOSHELL);            /* NOSHELL = /bin/false */

  /* 
   * 3. Set a secure umask.
   */
  (void) umask(ADMIN_UMASK);

} /* setEnvironment */

/*
 * 1. User must have a base gid of SYSADMIN_GID to use runas.
 * 2. ALLOW_GROUPIES: User can be a member (instead of having a base gid)
 *    of group SYSADMIN_GID to use runas.
 * 3. USE_UID_RANGE: User must have a uid higher than or including
 *    USE_RANGE_BOTTOM and lower than or including USE_RANGE_TOP
 *    to use runas.
 * 4. USE_ACF: If user cannot run any command as any user and/or group,
 *    then check the access file for authorzied commands to run.
 * 5. Access has been denied for the command.  Do not send a message to 
 *    stderr/syslogd and exit if the user only wants to know which commands 
 *    are available.
 *
 * Determines if "myuid" has permission to use runas
 *
 *
 * Sets global variable mygid.
 *
 * uid_t myuid                   Uid of user invoking runas         
 * uid_t myeui                   Effective uid of runas process     
 * gid_t mygid                   Gid of user invoking runas
 * int *argc                     Number of command line arguments   
 * char **argv                   Command line arguments             
 * char *runuidptr               Uid or username of user to run as  
 * char *rungidptr               Gid or groupname of user to run as 
 * BOOLEAN showCommands          Has --access been specified?
 * BOOLEAN needACF               Does the user need the ACF to use runas?
 *
 * input: argc, argv, myuid, myeuid, mygid, showCommands
 * output: argc, argv, runuid, rungid, needACF
 */
PRIVATE void checkAccess(uid_t myuid, 
			 uid_t myeuid, 
			 gid_t mygid, 
			 int *argc, 
			 char **argv, 
			 char *runuidptr, 
			 char *rungidptr, 
			 BOOLEAN showCommands, 
			 BOOLEAN *needACF)
{
  BOOLEAN noaccess = FALSE;        /* User can run any command by default. */
  char user[MAXUSERNAMESIZE];
  struct passwd *passwd;
#if ((ALLOW_GROUPIES == TRUE) || (USE_UID_RANGE == TRUE)) 
  BOOLEAN groupie = FALSE;
#endif
#if (ALLOW_GROUPIES == TRUE)
  struct group *gp;
  char **members;
#endif /* ALLOW_GROUPIES */

  passwd = getpwuid(myuid);
  if (passwd == NULL) {
    (void) esprintf(error, "%s: \"%s\", line %i getpwuid failed", label, __FILE__, __LINE__);
    (void) errorPMsg(error);
  }
  mygid = passwd->pw_gid;
  (void) bzero(user, MAXUSERNAMESIZE);
  (void) strncpy(user, passwd->pw_name, MAXUSERNAMELEN);

  /*
   * 1. User must have a base gid of SYSADMIN_GID to use runas.
   */
  if ((mygid != SYSADMIN_GID) && (myuid != 0)) {
    noaccess = TRUE;
  }  

  /*
   * 2. ALLOW_GROUPIES: User can be a member (instead of having a base gid)
   *    of group SYSADMIN_GID to use runas.
   */
#if (ALLOW_GROUPIES == TRUE)
  if (noaccess == TRUE) {
    /* User does not have a base gid of SYSADMIN_GID */
    gp = getgrgid((gid_t)SYSADMIN_GID);
    if (gp == NULL) { /* Not found or ENOMEM */
      (void) esprintf(error, "%s: \"%s\", line %i getgrgid failed", label, __FILE__, __LINE__);
      (void) errorPMsg(error);
    }
    members = gp->gr_mem;
    while ((*members) && (noaccess == TRUE)) {
      if (strncmp(*members, user, MAXUSERNAMELEN) == 0) {
	groupie = TRUE;
	noaccess = FALSE;
      }  /* if (strncmp) */
      members++;
    } /* while (*gp->gr_mem) */
  } /* if (noaccess = TRUE) */ 
#endif /* ALLOW_GROUPIES */

  /*
   * 3. USE_UID_RANGE: User must have a uid higher than or including
   *    USE_RANGE_BOTTOM and lower than or including USE_RANGE_TOP
   *    to use runas.
   */
#if (USE_UID_RANGE == TRUE)
  /* Sanity check for misconfigured systems. */
  if (UID_RANGE_TOP < UID_RANGE_BOTTOM) {
    (void) esprintf(error, "Uid range misconfigured.  Top uid: %lu, Bottom uid: %lu", (long unsigned)UID_RANGE_TOP, (long unsigned)UID_RANGE_BOTTOM);
    (void) errorMsg(error, FALSE);
  }
  if ((groupie == FALSE) && (((myuid > (uid_t)UID_RANGE_TOP) ||
      (myuid < (uid_t)UID_RANGE_BOTTOM)) && (myuid != 0))) {
    noaccess = TRUE;
  }
#endif /* USE_UID_RANGE */

  /*
   * 4. USE_ACF: If user cannot run any command as any user and/or group,
   *    then check the access file for authorzied commands to run.
   */
  if (noaccess == TRUE) {
    *needACF = TRUE;
    if (showCommands == FALSE) {
#if (USE_ACF == TRUE)
#if (TIGHT_CHECKING == TRUE)
    (void) security();    /* anal security check of access control file */
#endif /* TIGHT_CHECKING */
     /* read in the ACF entries and then check for access */
    if ((parseAcf(myuid, mygid) == 0) && (checkCommandAccess(argc, argv, runuidptr, rungidptr, user) == 0)) {
      noaccess = FALSE;
    } /* parseAcf() */
#endif /* USE_ACF */
    }
  }/* if (noaccess == TRUE) */

  /* 
   * 5. Access has been denied for the command.  Do not send a message to 
   *    stderr/syslogd and exit if the user only wants to know which commands 
   *    are available.
   */
  if ((noaccess == TRUE) && (showCommands == FALSE)) {
#if (USE_SYSLOG == TRUE)
    syslogCommandNOPRIV(argc, argv, myuid, myeuid, user); /* does not return */
#else /* print an error message if not using syslog */
    (void) errorMsg(NOLOG_NOPRIVMSG, FALSE); /* does not return */
#endif /* USE_SYSLOG */
  } /* if (noaccess == TRUE) && (showCommands == FALSE) */
} /* checkAccess */

/* 
 * infoVersion - Displays the following information:
 *
 * 1. Release version and date information.
 * 2. Copyright, warranty, and distribution information.
 */
PRIVATE void infoVersion(void)
{
  /*
   * 1. Release version and date information.
   */
  (void) printf("Runas %s (%s)\n", RELEASE_VERSION, RELEASE_DATE);

  /*
   * 2. Copyright, warranty, and distribution information.
   */
  (void) printf("Copyright 1998 by Chris Carpinello\n");
  (void) printf("Runas comes with NO WARRANTY, to the extent permitted by law.  You may\n");
  (void) printf("redistribute copies of Runas under the terms of the GNU General Public License.\n");
  (void) printf("For more information about these matters, see the file named COPYING provided\n");
  (void) printf("in the distribution.\n");
}

/* 
 * infoCompile - Displays the following information:
 *
 * 1. Default uid and gid.
 * 1. Conditional compile-time administrator-set options.
 * 3. Environment settings.
 */
PRIVATE void infoCompile(void)
{
#if (EXEC_ONLY_ADMIN_OWNED_FILES == TRUE)  
  char **okowners; 
#endif /* EXEC_ONLY_ADMIN_OWNED_FILES */

#if (ALLOW_ROOT_SHELLS == FALSE)
  char *rshell;
#endif /* ALLOW_ROOT_SHELLS */

/* 
 * This information is used to determine what configuration flags were set
 * at compilation time.  
 */
#if (USE_SYSLOG == TRUE)
  int i; /* Ok_Folks loop counter */
#endif

  /*
   * 1. Default uid and gid.
   */
  (void) printf("\nCommands are run as uid=%lu ", (long unsigned)DEFAULT_UID);
#ifdef ROOT_GID
  if (DEFAULT_UID == 0) {
    (void) printf("and gid=%lu ", (long unsigned)ROOT_GID);
  }
#endif /* ROOT_GID */
  (void) printf("by default.\n");

  /*
   * 2. Conditional compile-time administrator-set options.
   */
  (void) printf("\nConditional compile-time options:\n");

#if (USE_ACF == TRUE)
  /* ACF Netgroup support. */
  (void) printf("- %s\n", runas_netgroups);
  (void) printf("- %s", runas_tight);

  /* Anal access control file attribute checking */
#if (TIGHT_CHECKING == TRUE)
  (void) printf(":\n    Cannot be a link.\n");
  (void) printf("    Permission mode must be %o.\n", ACF_MODE);
  (void) printf("    Ownership of uid=%lu (%s) and gid=%lu (%s).", (long unsigned) ACF_UID, (getpwuid(ACF_UID)->pw_name), (long unsigned) ACF_GID, (getgrgid(ACF_GID)->gr_name));
#endif /* TIGHT_CHECKING */
  (void) printf("\n");

  /* Anal absolute command path access control file entry attribute checking */
  (void) printf("- %s", runas_tight2);
#if (TIGHT_CHECKING == TRUE)
  (void) printf(":\n    Cannot be a link.\n");
  (void) printf("    Cannot be group writable.\n");
  (void) printf("    Cannot be world writable.");
#endif /* TIGHT_CHECKING */
  (void) printf("\n");
#endif /* USE_ACF */

  /* USE_SYSLOG setting. */
  (void) printf("- %s", runas_syslog);
#if (USE_SYSLOG == TRUE)
  (void) printf(":  User ids exempted from logging:");

  /* Commands run by UIDs listed in Ok_Folks will bypass syslog(). */ 
  for (i = 0; Ok_Folks[i] != OK_FOLKS_DELIMITER; i++) {
    printf(" %lu", (long unsigned)Ok_Folks[i]);
  }
  if (i == 0) {
    (void) printf(" none");
  }
#endif /* USE_SYSLOG */
  (void) printf(".\n");

  (void) printf("- %s\n", runas_validate_user_group);

  (void) printf("- %s\n", runas_allow_root_shells);
#if (ALLOW_ROOT_SHELLS == FALSE)
  setusershell();
  while ((rshell = getusershell()) != NULL) {
    printf("  %s\n", rshell);
  }
  endusershell();
#endif /* ALLOW_ROOT_SHELLS */

#if (EXEC_ONLY_ADMIN_OWNED_FILES == TRUE)
  (void) printf("- %s", runas_exec_admin_files);
  okowners = ownedby;
  if (okowners == NULL) 
    (void) printf(" [none]"); /* pointless */
  else
    while (*okowners != NULL) {
      (void) printf(" %s", *okowners);
      okowners++;
    }
  (void) printf(".\n");
#endif /* EXEC_ONLY_ADMIN_OWNED_FILES */

#ifdef IGNORE_PREFIX_IN_PATH
  /* IGNORE_PREFIX_IN_PATH setting. */
  (void) printf("- %s\n    %s\n", runas_ignorepath, IGNORE_PREFIX_IN_PATH);
#endif /* IGNORE_PREFIX_IN_PATH */

#if (IGNORE_SYMLINKS == TRUE)
  /* IGNORE_SYMLINKS setting. */
  (void) printf("- %s\n", runas_symlinks);
#endif /* IGNORE_SYMLINKS */

  /*
   * 3. Environment settings.
   */
  (void) printf("- Environment settings:\n");

  /* Environment preservation. */
  if (preserve_env == TRUE)
    (void) printf("    %s", runas_preserve_env); 
  else 
    (void) printf("    %s", runas_sanitize_env); 
  if (preserve_env == PRESERVE_ENVIRONMENT_DEFAULT)
    (void) printf(" [default]");
  (void) printf("\n");
  /* UMASK setting. */
  (void) printf("    UMASK=%03lo (MODE=%3lo)\n", (long unsigned)ADMIN_UMASK, (long unsigned)(0666 & ~ADMIN_UMASK));
  /* SHELL setting. */  
  (void) printf("    SHELL=%s\n", NOSHELL);

  /* PATH setting. */
  (void) printf("    PATH=");
#if (USE_SECURE_PATH == TRUE)
  (void) printf("%s:", SECURE_PATH);
#endif /* USE_SECURE_PATH */
  (void) printf("$PATH\n      (%s)\n", runas_securepath);

} /* infoVersion */

/* 
 * infoBugReport
 * 
 * 1. Display user defined options.
 * 2. Display compiler information.
 * 3. Display compile-time system-configuration definitions.
 * 4. Display bug report information.
 *
 * uid_t myuid             uid of the administrator
 * gid_t mygid             gid of the administrator
 * BOOLEAN needACF         does the administrator depend on the ACF?
 */ 
PRIVATE void infoBugReport(uid_t myuid, 
			   gid_t mygid, 
			   BOOLEAN needACF)
{
  /* 
   * 1. Display user defined options.
   */
  (void) infoVersion();  
  (void) infoCompile();
  (void) printf("\n");
  (void) infoAccess(myuid, mygid, needACF);
  (void) printf("\n");

  /* 
   * 2. Display compiler information.
   */
  (void) printf("Compiled with %s ", COMPILE_CC); 
#ifdef __VERSION__
  (void) printf("(%s) ", __VERSION__);
#endif /* __VERSION__ */
#ifdef __TIME__ 
#ifdef __DATE__
  (void) printf("%s %s ", __TIME__, __DATE__);
#endif /* __DATE__ */
#endif /* __TIME__ */
  (void) printf("on a %s %i.%i", sysname, MAJORVER, MINORVER);
#ifdef BUGFIXVER
  (void) printf(".%i", BUGFIXVER);
#endif 
  (void) printf(" platform.\n");
  (void) printf("CFLAGS =%s\n", COMPILE_CFLAGS);
  (void) printf("LIBS =%s\n", COMPILE_LIBS);
  (void) printf("LDFLAGS =%s\n", COMPILE_LDFLAGS);

  /* 
   * 3. Display compile-time system-configuration definitions.
   */
  (void) printf("DEFINES =%s\n", COMPILE_DEFINES);

  /*
   * 4. Display bug report information.
   */
  (void) printf("\nReport bugs to %s.\n", MAINTAINER);
} /* infoBugReport */

/*
 * infoUsage
 *
 * 1. Display the usage information.
 */
PRIVATE void infoUsage(void)
{

/*
 * 1. Display the usage information.
 */
  (void) printf("%s", USAGE_MSG);   /* usage line */
  (void) printf("%s", USAGE_MSG1);  /* --preserve */
#if (PRESERVE_ENVIRONMENT_DEFAULT == TRUE)
  (void) printf(" [default]");
#endif /* PRESERVE_ENVIRONMENT_DEFAULT */
  (void) printf("\n%s", USAGE_MSG2); /* --sanitize */
#if (PRESERVE_ENVIRONMENT_DEFAULT == FALSE)
  (void) printf(" [default]");
#endif /* PRESERVE_ENVIRONMENT_DEFAULT */
  (void) printf("\n%s", USAGE_MSG3); /* the rest */
} /* infoUsage */

/*
 * infoAccess
 *
 * 1. If the user requesting access information does not rely on the access
 *    control file, then list the conditions allowing a user to use runas.
 * 2. Read in access control file entries.
 * 3. Display the elements of each access entry in the list.
 * 4. Free memory allocated by access list.
 */
PRIVATE void infoAccess(uid_t myuid, 
			gid_t mygid, 
			BOOLEAN needACF)
{
#if (USE_ACF == TRUE)
  accessentry *entry;            /* Access control file entry               */
  int commands = 0;              /* Number of commands a user has access to */
#endif /* USE_ACF */
  
  /*
   * 1. If the user requesting access information does not rely on the access
   * control file, then list the conditions allowing a user to use runas.
   */
  if (needACF == FALSE) {
    (void) printf("Users matching the following conditions can execute any command as any user \nand/or group:\n");

#if (USE_UID_RANGE == TRUE)
    /* UID range allowed to use runas. */
    (void) printf("  Must have a uid between and including %lu through %lu.\n", (long unsigned) UID_RANGE_BOTTOM, (long unsigned) UID_RANGE_TOP);
#endif /* USE_UID_RANGE */

    /* Must have gid=SYSADMIN_GID to use runas. */
    (void) printf("  Users must have a base gid=%lu", (long unsigned)SYSADMIN_GID);

#if (ALLOW_GROUPIES == TRUE)
    /* Groupies allowed to use runas. */
    (void) printf(" or belong to gid=%lu as a \"groupie\".\n", (long unsigned)SYSADMIN_GID);
#else
    (void) printf(".\n");    
#endif /* ALLOW_GROUPIES */
  } /* if needACF == FALSE */

#if (USE_ACF == TRUE)
  if (needACF == FALSE) {
    /* Access control file dependency. */
    (void) printf("%s commands available:\n", ACF_PATH);
    (void) printf(ACCESS_MSGDELIMITER);
  } /* if needACF == FALSE */

  /* 
   * 2. Read in access control file entries.
   */ 
  if (parseAcf(myuid, mygid) != 0) {
    (void) errorMsg("Error parsing access control file.", FALSE);
  }

  /*
   * 3. Display the elements of each access entry in the list.
   */
  entry = accesscf;
  while (entry != NULL) {
    /* Users that have authorization to run this command */ 
    if (needACF == FALSE) {
      (void) printAcfCommand(entry, needACF);
      if ((++commands > 0) && (entry->next != NULL)) {
	(void) printf(ACCESS_MSGDELIMITER);
      }
    } else {
      if (inAccessList(entry->users, getpwuid(myuid)->pw_name) == 0) {
	(void) printAcfCommand(entry, needACF);
	if ((++commands > 0) && (entry->next != NULL)) {
	  (void) printf(ACCESS_MSGDELIMITER);
	}
      }
    }
    entry = entry->next;
  }

  /*
   * 4. Free memory allocated by access list.
   */
  (void) freeAcfList();
  if ((needACF == TRUE) && (commands == 0)) {
    (void) errorMsg(NOLOG_NOPRIVMSG, FALSE);
  }
#else
  if (needACF == TRUE) {
    (void) errorMsg(NOLOG_NOPRIVMSG, FALSE);
  } else {
    (void) printf("The access control file is not in use.\n");
  }
#endif /* USE_ACF */
} /* infoAccess */

/*
 * parseUserGroupArg
 *
 * 1. Check boundary length of user.group argument to avoid any 
 *    possible buffer overflows.
 * 2. Parse the user and group to "run as" from the command line argument.
 * 3. Verify user is a valid user on the system.
 * 4. Verify group is a valid group on the system.
 *
 * char *arg;                            user.group parameter
 * uid_t myuid;                          uid of user invoking runas
 * char *runuidptr;                      username to run process as
 * char *rungidptr;                      groupname to run process as
 */
PRIVATE void parseUserGroupArg(char *arg, 
			       uid_t myuid, 
			       char *runuidptr, 
			       char *rungidptr)
{
  struct passwd *passwd = NULL;      /* passwd info of runuidptr            */
  struct group *group = NULL;        /* group info of rungidptr             */
  size_t ugidlen;                    /* length of user.group argv if given  */
  char *tmpstrtok;                   /* pointer to subsequent strtok() call */
  char *username;                    /* user's username                     */

  /* 
   * 1. Check boundary length of user.group argument to avoid any 
   *    possible buffer overflows.
   */
  ugidlen = strlen(arg);
  if (ugidlen > (MAXUSERNAMELEN + MAXGROUPNAMELEN + 2)) {
    (void) errorMsg("User or group name argument length out of bounds.", FALSE);
  }

  /* 
   * 2. Parse the user and group to "run as" from the command line argument.
   */
  arg++; /* Remove the "-" from the user.group argument.     */

  /* Neither user nor group specified. */
  if (--ugidlen < 1) { 
    (void) errorMsg("Empty user.group argument specified.", FALSE);
  }

  /* Only group specified. */
  if (strncmp(arg, ".", 1) == 0) { /* only group specified */
    arg++;
    passwd = getpwuid(myuid);
    username = passwd->pw_name;
    (void) strncpy(runuidptr, username, MAXUSERNAMELEN); /* keep same uid */
    if (--ugidlen < 1) {
      (void) errorMsg("Empty user.group argument specified.", FALSE);
    } else {
      (void) strncpy(rungidptr, arg, MAXGROUPNAMELEN);
    }
  /* User and group specified. */
  } else { 
    tmpstrtok=strtok(arg, ".");
    (void) strncpy(runuidptr, tmpstrtok, MAXUSERNAMELEN);
    tmpstrtok=strtok(NULL, ".");   
    /* NULL group specified. */
    if (tmpstrtok != NULL) {
      (void) strncpy(rungidptr, tmpstrtok, MAXGROUPNAMELEN);
    }
  }

  /* 
   * 3. Verify user is a valid user on the system.
   */
  (void) validateUser(runuidptr);
    
  /* 
   * 4. Verify group is a valid group on the system.
   */
  if ((rungidptr != NULL) && strlen(rungidptr)){  /* group arg specified */
    (void) validateGroup(rungidptr);
  }
  else { /* use the base group associated with runuidptr */
    passwd = getpwnam(runuidptr);
    group = getgrgid(passwd->pw_gid);
    if (group == NULL) { /* No such group or ENOMEM */
#if (VALIDATE_USER_GROUP == TRUE)
      if (errno == ENOMEM) { /* ENOMEM */
	(void) esprintf(error, "%s: \"%s\", line %i getgrgid failed for gid %lu", label, __FILE__, __LINE__, passwd->pw_gid);
	(void) errorPMsg(error); 
      } else { /* No such group */
	(void) fprintf(stderr, "%s: gid (%lu) associated with user (%s) does not have a group file entry.\n", label, (long unsigned)passwd->pw_gid, passwd->pw_name);
	(void) exitOK();
      }
#else
      (void) snprintf(rungidptr, MAXGROUPNAMELEN, "%lu", (long unsigned)passwd->pw_gid);
#endif /* VALIDATE_USER_GROUP */
    } else {
      (void) strncpy(rungidptr, group->gr_name, MAXGROUPNAMELEN);
    }
  }
} /* parseUserGroupArg */

/* 
 * Main 
 *
 *  1. USE_SYSLOG: Initialize syslog and syslog variables.
 *  2. Verify runas is running as root.
 *  3. Parse --sanitize, --preserve or --login flag.
 *  4. Determine if any part of user.group was specified as a command line 
 *     option.
 *  5. Display usage if less than 2 arguments specified or --help 
 *     argument was specified.
 *  6. Initialize user and group argument variable values to NULL. 
 *  7. Verify the user is allowed run commands via runas. 
 *  8. Process informational command line options and exit without
 *     starting a new process.
 *  9. runuidptr and rungidptr were not set in checkAccess, so user is
 *     allowed to run any command as any user and/or group.  Parse the 
 *     command line for the user and/or group to set the process id to.
 * 10. Set the real uid and gid of this process.  Send msg to syslog.
 * 11. Flush buffered streams.
 * 12. Execute the program with arguments and an environment. 
 * 13. Display an error and return the error value if execve() fails.
 *
 * int argc;                             Number of command line arguments
 * char *argv[];                         Command line arguments
 * char *envp[];                         Inherited environment
 */
int main(int argc, char **argv)
{
  uid_t myuid;                       /* uid of user invoking runas          */
  uid_t myeuid;                      /* effective uid of process            */
  gid_t mygid;                       /* gid of user invoking runas          */
  char runuidptr[MAXUSERNAMESIZE];   /* uid or username of user to run as   */
  char rungidptr[MAXGROUPNAMESIZE];  /* gid or groupname of group to run as */ 
  char *fullpath;                    /* fully qualified path of command     */
  BOOLEAN login_shell = FALSE;       /* has --login been specified?         */
  BOOLEAN setID = FALSE;             /* has user.group been specified?      */
  BOOLEAN showCommands = FALSE;      /* has --access been specified?        */
  BOOLEAN needACF = FALSE;           /* does user need ACF to use runas?    */

  myuid = getuid();                  /* uid of user invoking runas */
  myeuid = geteuid();                /* uid of effective user      */
  mygid = getgid();                  /* gid of user invoking ruans */

  (void) bzero(error, ERRORSIZE);    /* clear out the error buffer */

  /* 
   *  1. USE_SYSLOG: Initialize syslog and syslog variables.
   *
   * Note: This should happen before any calls to errorMsg() or errorPMsg()
   *       so error messages are handled properly.
   */
  label = strdup(argv[0]);
#if (USE_SYSLOG == TRUE)
  (void) initSyslog(label); 
#endif

  /*
   *  2. Verify runas is running as root.
   */
  if (myeuid != DEFAULT_UID) { 
    (void) errorMsg("runas is not setuid root", FALSE);
  }

  /*
   *  3. Parse environment flag: --sanitize, --preserve or --login
   */
  ++argv; /* increment argv[0] past the name runas was executed as. */
  while ((*argv != NULL) && (strncmp(*argv, "--", 2) == 0)) {
    if (strncmp(*argv, USAGE_OPTION_PRESERVE, sizeof(USAGE_OPTION_PRESERVE)) == 0) {
      preserve_env = TRUE;
      login_shell = FALSE;
      argc -= 1;
      /* --preserve flag set */
    } else if (strncmp(*argv, USAGE_OPTION_SANITIZE, sizeof(USAGE_OPTION_SANITIZE)) == 0) {
      /* --sanitize flag set */
      preserve_env = FALSE;
      login_shell = FALSE;
      argc -= 1;
    } else if (strncmp(*argv, USAGE_OPTION_LOGIN, sizeof(USAGE_OPTION_LOGIN)) == 0) {
      /* --login flag set */
      preserve_env = FALSE;
      login_shell = TRUE;
      argc -= 1;
    } else if (strncmp(*argv, USAGE_OPTION_WHICH, sizeof(USAGE_OPTION_WHICH)) == 0) {
      /* --which flag set */
      which_command = TRUE;
      argc -= 1;
    } else if (strncmp(*argv, USAGE_OPTION_ACCESS, sizeof(USAGE_OPTION_ACCESS)) == 0) {
      /* --access flag set */
      showCommands = TRUE;
      access_command = TRUE;
    } else if (strncmp(*argv, USAGE_OPTION_HELP, sizeof(USAGE_OPTION_HELP)) == 0) {
      /* --help flag set */
      help_command = TRUE;
    } else if (strncmp(*argv, USAGE_OPTION_VERSION, sizeof(USAGE_OPTION_VERSION)) == 0){
      /* --version flag set */
      version_command = TRUE;
    } else if (strncmp(*argv, USAGE_OPTION_INFO, sizeof(USAGE_OPTION_INFO)) == 0) {
      /* --info flag set */
      info_command = TRUE;
    } else {
      (void) esprintf(error, "Unrecognized option: %s", *argv);
      (void) errorMsg(error, FALSE);
    }
    ++argv;
  }

  /* 
   *  4. Determine if any part of user.group was specified as a command line 
   *     option.
   */
  if ((*argv != NULL) && (**argv == '-')) {
    setID = TRUE; /* explicitly set user.group command line option */
    (argc) -= 2;
  } else {
    (argc) -= 1; /* did not set user.group command line option */
  }

  /* 
   *  5. Display usage if less than 2 arguments specified or --help 
   *     argument was specified.
   */
  if ((argc < 1) ||  /* no command arguments were specified */
      (help_command)) { /* --help specified */
    (void) infoUsage();
    (void) exitOK();
  }
  
  /* 
   *  6. Initialize user and group argument variable values to NULL. 
   */
  (void) bzero(runuidptr, MAXUSERNAMESIZE);
  (void) bzero(rungidptr, MAXGROUPNAMESIZE); 

  /* 
   *  7. Verify the user is allowed run commands via runas. 
   */
  (void) checkAccess(myuid, myeuid, mygid, &argc, argv, runuidptr, rungidptr, showCommands, &needACF); 

  /* Force the environment to be sanitized for ACF users. */
  if (needACF == TRUE)
    preserve_env = FALSE;

  /* 
   *  8. Process informational command line options and exit without
   *     starting a new process.
   * 
   * Note: informational command line options contain very detailed 
   * information about how runas is configured and compiled.  
   * Since this sensitive information lists the commands and users allowed
   * to use runas, only display it if the user can already run any command 
   * as any user via runas.
   */
 
  /* Display access information. */
  if (access_command) {
    (void) infoAccess(myuid, mygid, needACF);
    (void) exitOK();
  }

  /* Display bugreport information. */
  if (info_command) {
    if (needACF == FALSE) {
      (void) infoBugReport(myuid, mygid, needACF);
      (void) exitOK();
    } else {
#if (USE_SYSLOG == TRUE)      
      (void) syslogCommandNOPRIV(&argc, argv, myuid, myeuid, NULL); /* does not return */
#else
      (void) errorMsg(NOLOG_NOPRIVMSG, FALSE); /* does not return */
#endif /* USE_SYSLOG */
    }
  }

  /* Display version information. */
  if (version_command) {
    if (needACF == FALSE) {
      (void) infoVersion();
      (void) exitOK();
    } else {
#if (USE_SYSLOG == TRUE) 
      (void) syslogCommandNOPRIV(&argc, argv, myuid, myeuid, NULL); /* does not return */
#else
      (void) errorMsg(NOLOG_NOPRIVMSG, FALSE); /* does not return */
#endif /* USE_SYSLOG */
    }
  }

  /* 
   *  9. runuidptr and rungidptr were not set in checkAccess, so user is
   *     allowed to run any command as any user and/or group.  Parse the 
   *     command line for the user and/or group to set the process id to.
   */
  if ((setID == TRUE) && (!strlen(runuidptr))) {
    (void) parseUserGroupArg(argv[0], myuid, runuidptr, rungidptr);
    argv++; /* Increment argv to get past user.group argument */
  }

  /* 
   * 10. Set the real uid and gid of this process.  Send msg to syslog.
   */
  (void) setIdentity(runuidptr, rungidptr, argc, argv, myuid, myeuid, setID, needACF);

  /*
   * 11. Flush buffered streams.
   *     USE_SYSLOG: Tidy up (close) syslog.
   */
  (void) fflush(stdout);
  (void) fflush(stderr);
#if (USE_SYSLOG == TRUE)
  (void) syslogCleanUp();
#endif /* USE_SYSLOG */

  if (! which_command) {
  /* 
   * 12. Execute the program with arguments and an environment. 
   */
    fullpath = strdup(argv[0]);
    argv[0] = basename(argv[0]);
    if ((login_shell) && (restricted_shell)) {
      /* prepend "-" to shell name to have the shell run as a login shell. */
      char *argv0;
      argv0 = __ecmalloc(strlen(argv[0]) + 2);
      argv0[0] = '-';
      strcpy(argv0 + 1, argv[0]);
      argv[0] = argv0;
    } 
    (void) execve(fullpath, argv, environ);
  
    /* 
     * 13. Display an error and return the error value if execve() fails.
     */
    (void) esprintf(error, "%s: %s", label, fullpath);
    (void) errorPMsg(error);
  }
  return(0);
} /* main */

/* end runas.c */
